"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AnalysisBase = void 0;
const PQP = require("@microsoft/powerquery-parser/lib/powerquery-parser");
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const trace_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/common/trace");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const inspection_1 = require("../inspection");
const __1 = require("..");
const scopeUtils_1 = require("../inspection/scope/scopeUtils");
const providers_1 = require("../providers");
const trace_2 = require("../trace");
// Implementation of Analysis.
// All public methods should:
//  - use ensureResultAsync to ensure a Result, allowing all protected methods to safetly throw.
//  - take an optional cancellation token.
class AnalysisBase {
    constructor(textDocument, analysisSettings) {
        this.textDocument = textDocument;
        // Local type cache that is shared across provider actions.
        this.typeCache = {
            scopeById: new Map(),
            typeById: new Map(),
        };
        this.initialCorrelationId = analysisSettings.initialCorrelationId;
        this.inspectionSettings = analysisSettings.inspectionSettings;
        this.locale = analysisSettings.inspectionSettings.locale;
        this.traceManager = analysisSettings.traceManager;
        this.languageAutocompleteItemProvider = analysisSettings.languageAutocompleteItemProviderFactory
            ? analysisSettings.languageAutocompleteItemProviderFactory(analysisSettings.inspectionSettings.locale)
            : new providers_1.LanguageAutocompleteItemProvider(analysisSettings.inspectionSettings.locale);
        this.libraryProvider = analysisSettings.libraryProviderFactory
            ? analysisSettings.libraryProviderFactory(analysisSettings.inspectionSettings.library, analysisSettings.inspectionSettings.locale)
            : new providers_1.LibraryProvider(analysisSettings.inspectionSettings.library, analysisSettings.inspectionSettings.locale);
        this.localDocumentProvider = analysisSettings.localDocumentProviderFactory
            ? analysisSettings.localDocumentProviderFactory(textDocument.uri.toString(), this.typeCache, analysisSettings.inspectionSettings.library, analysisSettings.inspectionSettings.locale)
            : new providers_1.LocalDocumentProvider(this.textDocument.uri.toString(), this.typeCache, analysisSettings.inspectionSettings.library, analysisSettings.inspectionSettings.locale);
        this.triedLexParse = this.tryLexParse();
    }
    getAutocompleteItems(position, cancellationTokenOverride) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getAutocompleteItems.name, this.initialCorrelationId);
            const context = yield this.getAutocompleteItemProviderContext(position, trace.id, cancellationTokenOverride);
            if (context === undefined) {
                trace.exit();
                return undefined;
            }
            const autocompleteItemTasks = [
                this.languageAutocompleteItemProvider.getAutocompleteItems(context),
                this.libraryProvider.getAutocompleteItems(context),
                this.localDocumentProvider.getAutocompleteItems(context),
            ];
            const autocompleteItems = [];
            // TODO: intellisense improvements
            // - honor expected data type
            for (const result of yield Promise.all(autocompleteItemTasks)) {
                if (powerquery_parser_1.ResultUtils.isOk(result) && result.value !== undefined) {
                    for (const item of result.value) {
                        autocompleteItems.push(item);
                    }
                }
            }
            trace.exit();
            return autocompleteItems;
        }), this.locale);
    }
    getDefinition(position, cancellationTokenOverride) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getDefinition.name, this.initialCorrelationId);
            const identifierContext = yield this.getDefinitionProviderContext(position, trace.id, cancellationTokenOverride);
            if (!identifierContext) {
                trace.exit();
                return undefined;
            }
            const result = yield this.localDocumentProvider.getDefinition(identifierContext);
            if (powerquery_parser_1.ResultUtils.isError(result)) {
                trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                throw result.error;
            }
            trace.exit();
            return result.value;
        }), this.locale);
    }
    getDocumentSymbols(cancellationTokenOverride) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const parseState = yield this.getParseStateOkOrThrow(undefined);
            if (parseState === undefined) {
                return undefined;
            }
            const nodeIdMapCollection = parseState.contextState.nodeIdMapCollection;
            const currentSymbols = [];
            const parentSymbolById = new Map();
            const cancellationToken = cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken;
            addIdentifierPairedExpressionSymbols(nodeIdMapCollection, currentSymbols, parentSymbolById, cancellationToken);
            addRecordSymbols(nodeIdMapCollection, currentSymbols, parentSymbolById, cancellationToken);
            return currentSymbols;
        }), this.locale);
    }
    getFoldingRanges(cancellationTokenOverride) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getFoldingRanges.name, this.initialCorrelationId);
            const parseState = yield this.getParseStateOkOrThrow(trace);
            if (parseState === undefined) {
                trace.exit();
                return undefined;
            }
            const result = yield this.localDocumentProvider.getFoldingRanges({
                traceManager: this.traceManager,
                initialCorrelationId: trace.id,
                cancellationToken: cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken,
                nodeIdMapCollection: parseState.contextState.nodeIdMapCollection,
            });
            if (powerquery_parser_1.ResultUtils.isError(result)) {
                trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                throw result.error;
            }
            trace.exit();
            return result.value;
        }), this.locale);
    }
    getHover(position, cancellationTokenOverride) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getHover.name, this.initialCorrelationId);
            const hoverProviderContext = yield this.getHoverProviderContext(position, trace.id, cancellationTokenOverride);
            if (!hoverProviderContext) {
                trace.exit();
                return undefined;
            }
            const triedHovers = yield Promise.all([
                this.localDocumentProvider.getHover(hoverProviderContext),
                this.libraryProvider.getHover(hoverProviderContext),
            ]);
            // Keep track if any error was thrown, prioritized by the order of the providers
            let firstError;
            for (const triedHover of triedHovers) {
                if (powerquery_parser_1.ResultUtils.isOk(triedHover)) {
                    if (triedHover.value !== undefined) {
                        trace.exit();
                        return triedHover.value;
                    }
                }
                else if (firstError === undefined) {
                    firstError = triedHover.error;
                }
            }
            if (firstError !== undefined) {
                trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                throw firstError;
            }
            trace.exit();
            return undefined;
        }), this.locale);
    }
    // We should only get an undefined for an activeNode iff a parse pass hasn't been done.
    getActiveNode(position) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const parseState = yield this.getParseStateOkOrThrow(undefined);
            if (parseState === undefined) {
                return undefined;
            }
            return inspection_1.ActiveNodeUtils.activeNode(parseState.contextState.nodeIdMapCollection, position);
        }), this.locale);
    }
    // undefined means either the parse pass wasn't done (lexing error), or there was no error.
    getParseError() {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const triedLexParse = yield this.triedLexParse;
            if (PQP.TaskUtils.isParseStageCommonError(triedLexParse)) {
                throw triedLexParse.error;
            }
            else if (PQP.TaskUtils.isParseStageError(triedLexParse)) {
                return triedLexParse.error;
            }
            else {
                return undefined;
            }
        }), this.locale);
    }
    // We should only get an undefined for an activeNode iff a parse pass hasn't been done.
    getParseState() {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const triedLexParse = yield this.triedLexParse;
            if (PQP.TaskUtils.isParseStageCommonError(triedLexParse)) {
                throw triedLexParse.error;
            }
            else if (PQP.TaskUtils.isParseStageError(triedLexParse)) {
                return triedLexParse.parseState;
            }
            else if (PQP.TaskUtils.isParseStageOk(triedLexParse)) {
                return triedLexParse.parseState;
            }
            else {
                return undefined;
            }
        }), this.locale);
    }
    getPartialSemanticTokens(cancellationTokenOverride) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getPartialSemanticTokens.name, this.initialCorrelationId);
            const parseState = yield this.getParseStateOkOrThrow(trace);
            if (parseState === undefined) {
                trace.exit();
                return undefined;
            }
            const result = yield this.localDocumentProvider.getPartialSemanticTokens({
                traceManager: this.traceManager,
                initialCorrelationId: trace.id,
                cancellationToken: cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken,
                library: this.inspectionSettings.library,
                parseState,
            });
            if (powerquery_parser_1.ResultUtils.isError(result)) {
                trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                throw result.error;
            }
            trace.exit();
            return result.value;
        }), this.locale);
    }
    getSignatureHelp(position, cancellationTokenOverride) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getSignatureHelp.name, this.initialCorrelationId);
            const signatureProviderContext = yield this.getSignatureProviderContext(position, trace.id, cancellationTokenOverride);
            if (signatureProviderContext === undefined) {
                trace.exit();
                return undefined;
            }
            const signatureTasks = yield Promise.all([
                this.localDocumentProvider.getSignatureHelp(signatureProviderContext),
                this.libraryProvider.getSignatureHelp(signatureProviderContext),
            ]);
            for (const task of signatureTasks) {
                if (powerquery_parser_1.ResultUtils.isOk(task) && task.value !== undefined) {
                    trace.exit();
                    return task.value;
                }
            }
            trace.exit();
            return undefined;
        }), this.locale);
    }
    getRenameEdits(position, newName, cancellationTokenOverride) {
        return powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getRenameEdits.name, this.initialCorrelationId);
            const parseState = yield this.getParseStateOkOrThrow(trace);
            const activeNode = yield this.getActiveNodeOkOrThrow(position, trace);
            if (parseState === undefined ||
                activeNode === undefined ||
                !inspection_1.ActiveNodeUtils.isPositionInBounds(activeNode) ||
                activeNode.inclusiveIdentifierUnderPosition === undefined) {
                trace.exit();
                return undefined;
            }
            void (yield this.inspectNodeScope(activeNode, trace.id, cancellationTokenOverride));
            const leafIdentifier = activeNode.inclusiveIdentifierUnderPosition;
            const scopeById = this.typeCache.scopeById;
            const identifiersToBeEdited = [];
            let valueCreator = undefined;
            if (leafIdentifier.node.kind === language_1.Ast.NodeKind.GeneralizedIdentifier) {
                valueCreator = leafIdentifier.node;
            }
            else if ((leafIdentifier.node.kind === language_1.Ast.NodeKind.Identifier ||
                leafIdentifier.node.kind === language_1.Ast.NodeKind.IdentifierExpression) &&
                scopeById) {
                // need to find this key value and modify all referring it
                const identifierExpression = leafIdentifier.node.kind === language_1.Ast.NodeKind.IdentifierExpression
                    ? leafIdentifier.node.identifier
                    : leafIdentifier.node;
                switch (identifierExpression.identifierContextKind) {
                    case language_1.Ast.IdentifierContextKind.Key:
                    case language_1.Ast.IdentifierContextKind.Parameter:
                        // it is the identifier creating the value
                        valueCreator = identifierExpression;
                        break;
                    case language_1.Ast.IdentifierContextKind.Value: {
                        // it is the identifier referring the value
                        let nodeScope = scopeById.get(identifierExpression.id);
                        // there might be a chance that its scope did not get populated yet, do another try
                        if (!nodeScope) {
                            void (yield __1.Inspection.tryNodeScope(this.inspectionSettings, parseState.contextState.nodeIdMapCollection, this.inspectionSettings.eachScopeById, identifierExpression.id, scopeById));
                            nodeScope = scopeById.get(identifierExpression.id);
                        }
                        const scopeItem = (0, scopeUtils_1.findScopeItemByLiteral)(nodeScope, leafIdentifier.normalizedLiteral);
                        if (scopeItem) {
                            const valueCreatorSearch = (0, scopeUtils_1.scopeCreatorIdentifier)(scopeItem);
                            if ((valueCreatorSearch === null || valueCreatorSearch === void 0 ? void 0 : valueCreatorSearch.kind) === language_1.Ast.NodeKind.Identifier) {
                                if (valueCreatorSearch.identifierContextKind === language_1.Ast.IdentifierContextKind.Key ||
                                    valueCreatorSearch.identifierContextKind === language_1.Ast.IdentifierContextKind.Parameter) {
                                    valueCreator = valueCreatorSearch;
                                }
                                else {
                                    identifiersToBeEdited.push(valueCreatorSearch);
                                }
                            }
                            else if ((valueCreatorSearch === null || valueCreatorSearch === void 0 ? void 0 : valueCreatorSearch.kind) === language_1.Ast.NodeKind.GeneralizedIdentifier) {
                                valueCreator = valueCreatorSearch;
                            }
                            else if (valueCreatorSearch) {
                                identifiersToBeEdited.push(valueCreatorSearch);
                            }
                        }
                        break;
                    }
                    case language_1.Ast.IdentifierContextKind.Keyword:
                    default:
                        // only modify once
                        identifiersToBeEdited.push(identifierExpression);
                        break;
                }
            }
            if (valueCreator) {
                // need to populate the other identifiers referring it
                identifiersToBeEdited.push(...(yield this.collectAllIdentifiersBeneath(parseState.contextState.nodeIdMapCollection, valueCreator)));
            }
            // if none found, directly put inclusiveIdentifierUnderPosition in if it exists
            if (identifiersToBeEdited.length === 0 && leafIdentifier) {
                identifiersToBeEdited.push(leafIdentifier.node.kind === language_1.Ast.NodeKind.IdentifierExpression
                    ? leafIdentifier.node.identifier
                    : leafIdentifier.node);
            }
            const result = identifiersToBeEdited.map((one) => ({
                range: __1.PositionUtils.rangeFromTokenRange(one.tokenRange),
                newText: newName,
            }));
            trace.exit();
            return result;
        }), this.locale);
    }
    dispose() {
        this.typeCache.scopeById.clear();
        this.typeCache.typeById.clear();
    }
    static isValidHoverIdentifier(activeNode) {
        var _a;
        const ancestry = activeNode.ancestry;
        const parent = ancestry[1];
        if (((_a = parent === null || parent === void 0 ? void 0 : parent.node) === null || _a === void 0 ? void 0 : _a.kind) === language_1.Ast.NodeKind.Parameter) {
            return false;
        }
        return true;
    }
    getAutocompleteItemProviderContext(position, correlationId, cancellationTokenOverride) {
        return __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getAutocompleteItemProviderContext.name, correlationId);
            const activeNode = yield this.getActiveNodeOkOrThrow(position, trace);
            if (activeNode === undefined || (inspection_1.ActiveNodeUtils.isPositionInBounds(activeNode) && activeNode.isInKey)) {
                trace.exit();
                return undefined;
            }
            const autocomplete = powerquery_parser_1.Assert.asDefined(yield this.inspectAutocomplete(activeNode, trace.id, cancellationTokenOverride), "we have a truthy activeNode meaning we have a truthy parseState");
            const triedNodeScope = powerquery_parser_1.Assert.asDefined(yield this.inspectNodeScope(activeNode, trace.id, cancellationTokenOverride), "we have a truthy activeNode meaning we have a truthy parseState");
            const triedScopeType = powerquery_parser_1.Assert.asDefined(yield this.inspectScopeType(activeNode, trace.id, cancellationTokenOverride), "we have a truthy activeNode meaning we have a truthy parseState");
            const activeLeafIdentifier = inspection_1.ActiveNodeUtils.isPositionInBounds(activeNode)
                ? activeNode.inclusiveIdentifierUnderPosition
                : undefined;
            let result;
            const cancellationToken = cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken;
            if (activeLeafIdentifier !== undefined) {
                result = {
                    autocomplete,
                    triedNodeScope,
                    triedScopeType,
                    traceManager: this.traceManager,
                    cancellationToken,
                    initialCorrelationId: trace.id,
                    range: __1.CommonTypesUtils.rangeFromTokenRange(activeLeafIdentifier.node.tokenRange),
                    text: activeLeafIdentifier.normalizedLiteral,
                    tokenKind: activeLeafIdentifier.node.kind,
                };
            }
            else {
                result = {
                    autocomplete,
                    triedNodeScope,
                    triedScopeType,
                    traceManager: this.traceManager,
                    cancellationToken,
                    initialCorrelationId: trace.id,
                };
            }
            trace.exit();
            return result;
        });
    }
    getDefinitionProviderContext(position, correlationId, cancellationTokenOverride) {
        return __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getDefinitionProviderContext.name, correlationId);
            const activeNode = yield this.getActiveNodeOkOrThrow(position, trace);
            if (activeNode === undefined ||
                !inspection_1.ActiveNodeUtils.isPositionInBounds(activeNode) ||
                activeNode.exclusiveIdentifierUnderPosition === undefined ||
                !AnalysisBase.isValidHoverIdentifier(activeNode)) {
                trace.exit();
                return undefined;
            }
            const identifierUnderPosition = activeNode.exclusiveIdentifierUnderPosition;
            const identifier = identifierUnderPosition.node.kind === language_1.Ast.NodeKind.IdentifierExpression
                ? identifierUnderPosition.node.identifier
                : identifierUnderPosition.node;
            const cancellationToken = cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken;
            const context = {
                traceManager: this.traceManager,
                range: __1.CommonTypesUtils.rangeFromTokenRange(identifier.tokenRange),
                initialCorrelationId: trace.id,
                cancellationToken,
                identifier,
                triedNodeScope: powerquery_parser_1.Assert.asDefined(yield this.inspectNodeScope(activeNode, trace.id, cancellationToken), "we got a truthy activeNode meaning we have a truthy parseState"),
            };
            trace.exit();
            return context;
        });
    }
    getHoverProviderContext(position, correlationId, cancellationTokenOverride) {
        return __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getHoverProviderContext.name, correlationId);
            const parseState = yield this.getParseStateOkOrThrow(trace);
            const activeNode = yield this.getActiveNodeOkOrThrow(position, trace);
            if (parseState === undefined || activeNode === undefined || !inspection_1.ActiveNodeUtils.isPositionInBounds(activeNode)) {
                trace.exit();
                return undefined;
            }
            const identifierUnderPosition = activeNode.exclusiveIdentifierUnderPosition;
            if (identifierUnderPosition === undefined || !AnalysisBase.isValidHoverIdentifier(activeNode)) {
                trace.exit();
                return undefined;
            }
            const cancellationToken = cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken;
            const identifier = identifierUnderPosition.node.kind === language_1.Ast.NodeKind.IdentifierExpression
                ? identifierUnderPosition.node.identifier
                : identifierUnderPosition.node;
            const context = {
                traceManager: this.traceManager,
                range: __1.CommonTypesUtils.rangeFromTokenRange(identifier.tokenRange),
                identifier,
                cancellationToken,
                initialCorrelationId: trace.id,
                activeNode,
                inspectionSettings: Object.assign(Object.assign({}, this.inspectionSettings), { cancellationToken, initialCorrelationId: correlationId }),
                parseState,
                triedNodeScope: powerquery_parser_1.Assert.asDefined(yield this.inspectNodeScope(activeNode, trace.id, cancellationToken), "we have a truthy activeNode meaning we have a truthy parseState"),
                triedScopeType: powerquery_parser_1.Assert.asDefined(yield this.inspectScopeType(activeNode, trace.id, cancellationToken), "we have a truthy activeNode meaning we have a truthy parseState"),
            };
            trace.exit();
            return context;
        });
    }
    getSignatureProviderContext(position, correlationId, cancellationTokenOverride) {
        return __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.getSignatureProviderContext.name, correlationId);
            const parseState = yield this.getParseStateOkOrThrow(trace);
            const activeNode = yield this.getActiveNodeOkOrThrow(position, trace);
            const triedCurrentInvokeExpression = yield this.inspectCurrentInvokeExpression(position, trace.id, cancellationTokenOverride);
            if (parseState === undefined ||
                activeNode === undefined ||
                triedCurrentInvokeExpression === undefined ||
                powerquery_parser_1.ResultUtils.isError(triedCurrentInvokeExpression) ||
                triedCurrentInvokeExpression.value === undefined) {
                trace.exit();
                return undefined;
            }
            const invokeExpression = triedCurrentInvokeExpression.value;
            const functionName = invokeExpression.name !== undefined ? invokeExpression.name : undefined;
            const argumentOrdinal = invokeExpression.arguments !== undefined ? invokeExpression.arguments.argumentOrdinal : undefined;
            if (functionName === undefined || argumentOrdinal === undefined) {
                trace.exit();
                return undefined;
            }
            const cancellationToken = cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken;
            const result = {
                argumentOrdinal,
                functionName,
                isNameInLocalScope: invokeExpression.isNameInLocalScope,
                functionType: invokeExpression.functionType,
                traceManager: this.traceManager,
                cancellationToken,
                initialCorrelationId: trace.id,
                triedCurrentInvokeExpression: yield __1.Inspection.tryCurrentInvokeExpression(Object.assign(Object.assign({}, this.inspectionSettings), { cancellationToken, initialCorrelationId: trace.id }), parseState.contextState.nodeIdMapCollection, activeNode, this.typeCache),
            };
            trace.exit();
            return result;
        });
    }
    getTypeCache() {
        return this.typeCache;
    }
    collectAllIdentifiersBeneath(nodeIdMapCollection, valueCreator) {
        var _a, _b, _c, _d, _e;
        return __awaiter(this, void 0, void 0, function* () {
            const astNodeById = nodeIdMapCollection.astNodeById;
            const childIdsById = nodeIdMapCollection.childIdsById;
            const originLiteral = valueCreator.literal;
            const entry = (0, scopeUtils_1.findDirectUpperScopeExpression)(nodeIdMapCollection, valueCreator.id);
            if (entry) {
                const allIdentifierIdSet = (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.Identifier)) !== null && _a !== void 0 ? _a : new Set();
                const allGeneralizedIdentifierIdSet = (_b = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.GeneralizedIdentifier)) !== null && _b !== void 0 ? _b : new Set();
                const current = entry;
                const idsByTiers = [];
                let currentTier = ((_c = childIdsById.get(current.id)) !== null && _c !== void 0 ? _c : []).slice();
                while (currentTier.length) {
                    const toBePushed = currentTier.filter((one) => allIdentifierIdSet.has(one) || allGeneralizedIdentifierIdSet.has(one));
                    toBePushed.length && idsByTiers.push(toBePushed);
                    currentTier = currentTier.slice();
                    let nextTier = [];
                    while (currentTier.length) {
                        const oneNode = (_d = currentTier.shift()) !== null && _d !== void 0 ? _d : -1;
                        const childrenOfTheNode = ((_e = childIdsById.get(oneNode)) !== null && _e !== void 0 ? _e : []).slice();
                        nextTier = nextTier.concat(childrenOfTheNode);
                    }
                    currentTier = nextTier;
                }
                // filter literal names
                const completeIdentifierNodes = idsByTiers
                    .flat(1)
                    .map((one) => astNodeById.get(one))
                    .filter(Boolean);
                let filteredIdentifierNodes = completeIdentifierNodes.filter((one) => one.literal === originLiteral);
                // populate the scope items for each
                yield Promise.all(filteredIdentifierNodes
                    .filter((oneIdentifier) => oneIdentifier.kind === language_1.Ast.NodeKind.GeneralizedIdentifier ||
                    oneIdentifier.identifierContextKind === language_1.Ast.IdentifierContextKind.Value)
                    .map((oneIdentifier) => __1.Inspection.tryNodeScope(this.inspectionSettings, nodeIdMapCollection, this.inspectionSettings.eachScopeById, oneIdentifier.id, this.typeCache.scopeById)));
                filteredIdentifierNodes = filteredIdentifierNodes.filter((oneIdentifierNode) => {
                    if (oneIdentifierNode.kind === language_1.Ast.NodeKind.GeneralizedIdentifier) {
                        return oneIdentifierNode.id === valueCreator.id;
                    }
                    else if (oneIdentifierNode.identifierContextKind === language_1.Ast.IdentifierContextKind.Value) {
                        const theScope = this.typeCache.scopeById.get(oneIdentifierNode.id);
                        const theScopeItem = (0, scopeUtils_1.findScopeItemByLiteral)(theScope, originLiteral);
                        const theCreatorIdentifier = (0, scopeUtils_1.scopeCreatorIdentifier)(theScopeItem);
                        return theCreatorIdentifier && theCreatorIdentifier.id === valueCreator.id;
                    }
                    else if (oneIdentifierNode.identifierContextKind === language_1.Ast.IdentifierContextKind.Key ||
                        oneIdentifierNode.identifierContextKind === language_1.Ast.IdentifierContextKind.Parameter) {
                        return oneIdentifierNode.id === valueCreator.id;
                    }
                    return false;
                });
                return filteredIdentifierNodes;
            }
            return [];
        });
    }
    // Performs a lex + parse of the document, then caches the results into local variables.
    // Any code which operates on the parse state should use this method to ensure that the parse state is initialized.
    tryLexParse() {
        return __awaiter(this, void 0, void 0, function* () {
            const trace = this.traceManager.entry(trace_2.ValidationTraceConstant.AnalysisBase, this.tryLexParse.name, this.initialCorrelationId);
            const updatedSettings = Object.assign(Object.assign({}, this.inspectionSettings), { initialCorrelationId: trace.id });
            const triedLexParse = yield PQP.TaskUtils.tryLexParse(updatedSettings, this.textDocument.getText());
            trace.exit();
            return triedLexParse;
        });
    }
    inspectAutocomplete(activeNode, correlationId, cancellationTokenOverride) {
        return __awaiter(this, void 0, void 0, function* () {
            const parseState = yield this.getParseStateOkOrThrow(undefined);
            const parseError = yield this.getParseErrorOkOrThrow(undefined);
            if (parseState === undefined) {
                return undefined;
            }
            return yield __1.Inspection.autocomplete(Object.assign(Object.assign({}, this.inspectionSettings), { initialCorrelationId: correlationId, cancellationToken: cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken }), parseState, this.typeCache, activeNode, parseError);
        });
    }
    inspectCurrentInvokeExpression(position, correlationId, cancellationTokenOverride) {
        return __awaiter(this, void 0, void 0, function* () {
            const parseState = yield this.getParseStateOkOrThrow(undefined);
            // parseState needs to be ok, and must be truthy.
            const activeNode = yield this.getActiveNodeOkOrThrow(position, undefined);
            if (parseState === undefined || activeNode === undefined) {
                return undefined;
            }
            return yield __1.Inspection.tryCurrentInvokeExpression(Object.assign(Object.assign({}, this.inspectionSettings), { initialCorrelationId: correlationId, cancellationToken: cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken }), parseState.contextState.nodeIdMapCollection, activeNode, this.typeCache);
        });
    }
    inspectNodeScope(activeNode, correlationId, cancellationTokenOverride) {
        return __awaiter(this, void 0, void 0, function* () {
            const parseState = yield this.getParseStateOkOrThrow(undefined);
            if (parseState === undefined || !inspection_1.ActiveNodeUtils.isPositionInBounds(activeNode)) {
                return undefined;
            }
            return yield __1.Inspection.tryNodeScope(Object.assign(Object.assign({}, this.inspectionSettings), { cancellationToken: cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken, initialCorrelationId: correlationId }), parseState.contextState.nodeIdMapCollection, this.inspectionSettings.eachScopeById, inspection_1.ActiveNodeUtils.assertGetLeaf(activeNode).node.id, this.typeCache.scopeById);
        });
    }
    inspectScopeType(activeNode, correlationId, cancellationTokenOverride) {
        return __awaiter(this, void 0, void 0, function* () {
            const parseState = yield this.getParseStateOkOrThrow(undefined);
            if (parseState === undefined || !inspection_1.ActiveNodeUtils.isPositionInBounds(activeNode)) {
                return undefined;
            }
            return yield __1.Inspection.tryScopeType(Object.assign(Object.assign({}, this.inspectionSettings), { cancellationToken: cancellationTokenOverride !== null && cancellationTokenOverride !== void 0 ? cancellationTokenOverride : this.inspectionSettings.cancellationToken, initialCorrelationId: correlationId }), parseState.contextState.nodeIdMapCollection, inspection_1.ActiveNodeUtils.assertGetLeaf(activeNode).node.id, this.typeCache);
        });
    }
    getActiveNodeOkOrThrow(position, trace) {
        return __awaiter(this, void 0, void 0, function* () {
            const activeNodeResult = yield this.getActiveNode(position);
            if (powerquery_parser_1.ResultUtils.isError(activeNodeResult)) {
                trace === null || trace === void 0 ? void 0 : trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                throw activeNodeResult.error;
            }
            return activeNodeResult.value;
        });
    }
    getParseErrorOkOrThrow(trace) {
        return __awaiter(this, void 0, void 0, function* () {
            const parseErrorResult = yield this.getParseError();
            if (powerquery_parser_1.ResultUtils.isError(parseErrorResult)) {
                trace === null || trace === void 0 ? void 0 : trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                throw parseErrorResult.error;
            }
            return parseErrorResult.value;
        });
    }
    getParseStateOkOrThrow(trace) {
        return __awaiter(this, void 0, void 0, function* () {
            const parseStateResult = yield this.getParseState();
            if (powerquery_parser_1.ResultUtils.isError(parseStateResult)) {
                trace === null || trace === void 0 ? void 0 : trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                throw parseStateResult.error;
            }
            return parseStateResult.value;
        });
    }
    applyAnalysisSettings(analysisSettings) {
        this.initialCorrelationId = analysisSettings.initialCorrelationId;
        this.inspectionSettings = analysisSettings.inspectionSettings;
        this.locale = analysisSettings.inspectionSettings.locale;
        this.traceManager = analysisSettings.traceManager;
    }
}
exports.AnalysisBase = AnalysisBase;
function addIdentifierPairedExpressionSymbols(nodeIdMapCollection, currentSymbols, parentSymbolById, cancellationToken) {
    var _a;
    const identifierPairedExpressionIds = (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.IdentifierPairedExpression)) !== null && _a !== void 0 ? _a : new Set();
    for (const nodeId of identifierPairedExpressionIds) {
        cancellationToken === null || cancellationToken === void 0 ? void 0 : cancellationToken.throwIfCancelled();
        const xorNode = parser_1.NodeIdMapUtils.assertXorChecked(nodeIdMapCollection, nodeId, language_1.Ast.NodeKind.IdentifierPairedExpression);
        if (!parser_1.XorNodeUtils.isAst(xorNode)) {
            continue;
        }
        const astNode = xorNode.node;
        const documentSymbol = __1.InspectionUtils.getSymbolForIdentifierPairedExpression(astNode);
        addDocumentSymbols(nodeIdMapCollection, parentSymbolById, nodeId, currentSymbols, documentSymbol);
        parentSymbolById.set(nodeId, documentSymbol);
    }
}
function addRecordSymbols(nodeIdMapCollection, currentSymbols, parentSymbolById, cancellationToken) {
    var _a, _b;
    const recordIdCollections = [
        (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.RecordExpression)) !== null && _a !== void 0 ? _a : new Set(),
        (_b = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.RecordLiteral)) !== null && _b !== void 0 ? _b : new Set(),
    ];
    for (const collection of recordIdCollections) {
        for (const nodeId of collection) {
            cancellationToken === null || cancellationToken === void 0 ? void 0 : cancellationToken.throwIfCancelled();
            const xorNode = parser_1.NodeIdMapUtils.assertXorChecked(nodeIdMapCollection, nodeId, [language_1.Ast.NodeKind.RecordExpression, language_1.Ast.NodeKind.RecordLiteral]);
            if (!parser_1.XorNodeUtils.isAst(xorNode)) {
                continue;
            }
            const astNode = xorNode.node;
            // Process the record if the immediate parent is a Struct
            const parentId = nodeIdMapCollection.parentIdById.get(astNode.id);
            const parentSymbol = parentId ? parentSymbolById.get(parentId) : undefined;
            if (parentSymbol && parentSymbol.kind === vscode_languageserver_types_1.SymbolKind.Struct) {
                const fieldSymbols = __1.InspectionUtils.getSymbolsForRecord(astNode);
                if (fieldSymbols.length > 0) {
                    addDocumentSymbols(nodeIdMapCollection, parentSymbolById, astNode.id, currentSymbols, ...fieldSymbols);
                }
            }
        }
    }
}
function addDocumentSymbols(nodeIdMapCollection, parentSymbolById, nodeId, currentSymbols, ...newSymbols) {
    const parentSymbol = findParentSymbol(nodeIdMapCollection, parentSymbolById, nodeId);
    if (parentSymbol) {
        if (!parentSymbol.children) {
            parentSymbol.children = [];
        }
        parentSymbol.children.push(...newSymbols);
        return;
    }
    // Add to the top level
    currentSymbols.push(...newSymbols);
}
function findParentSymbol(nodeIdMapCollection, parentSymbolById, nodeId) {
    // Get parent for current node
    const parentNodeId = nodeIdMapCollection.parentIdById.get(nodeId);
    if (!parentNodeId) {
        // No more parents to check
        return undefined;
    }
    let parentSymbol = parentSymbolById.get(parentNodeId);
    if (!parentSymbol) {
        parentSymbol = findParentSymbol(nodeIdMapCollection, parentSymbolById, parentNodeId);
    }
    return parentSymbol;
}
//# sourceMappingURL=analysisBase.js.map