"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDocumentSymbols = void 0;
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const commonTypes_1 = require("./commonTypes");
const _1 = require(".");
function getDocumentSymbols(nodeIdMapCollection, cancellationToken) {
    const currentSymbols = [];
    const parentSymbolById = new Map();
    addIdentifierPairedExpressionSymbols(nodeIdMapCollection, currentSymbols, parentSymbolById, cancellationToken);
    addRecordSymbols(nodeIdMapCollection, currentSymbols, parentSymbolById, cancellationToken);
    return currentSymbols;
}
exports.getDocumentSymbols = getDocumentSymbols;
function addIdentifierPairedExpressionSymbols(nodeIdMapCollection, currentSymbols, parentSymbolById, cancellationToken) {
    var _a;
    const identifierPairedExpressionIds = (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.IdentifierPairedExpression)) !== null && _a !== void 0 ? _a : new Set();
    for (const nodeId of identifierPairedExpressionIds) {
        cancellationToken.throwIfCancelled();
        const xorNode = parser_1.NodeIdMapUtils.assertXorChecked(nodeIdMapCollection, nodeId, language_1.Ast.NodeKind.IdentifierPairedExpression);
        if (!parser_1.XorNodeUtils.isAst(xorNode)) {
            continue;
        }
        const astNode = xorNode.node;
        const documentSymbol = _1.InspectionUtils.getSymbolForIdentifierPairedExpression(astNode);
        addDocumentSymbols(nodeIdMapCollection, parentSymbolById, nodeId, currentSymbols, documentSymbol);
        parentSymbolById.set(nodeId, documentSymbol);
    }
}
function addRecordSymbols(nodeIdMapCollection, currentSymbols, parentSymbolById, cancellationToken) {
    var _a, _b;
    const recordIdCollections = [
        (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.RecordExpression)) !== null && _a !== void 0 ? _a : new Set(),
        (_b = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.RecordLiteral)) !== null && _b !== void 0 ? _b : new Set(),
    ];
    for (const collection of recordIdCollections) {
        for (const nodeId of collection) {
            cancellationToken.throwIfCancelled();
            const xorNode = parser_1.NodeIdMapUtils.assertXorChecked(nodeIdMapCollection, nodeId, [language_1.Ast.NodeKind.RecordExpression, language_1.Ast.NodeKind.RecordLiteral]);
            if (!parser_1.XorNodeUtils.isAst(xorNode)) {
                continue;
            }
            const astNode = xorNode.node;
            // Process the record if the immediate parent is a Struct
            const parentId = nodeIdMapCollection.parentIdById.get(astNode.id);
            const parentSymbol = parentId ? parentSymbolById.get(parentId) : undefined;
            if (parentSymbol && parentSymbol.kind === commonTypes_1.SymbolKind.Struct) {
                const fieldSymbols = _1.InspectionUtils.getSymbolsForRecord(astNode);
                if (fieldSymbols.length > 0) {
                    addDocumentSymbols(nodeIdMapCollection, parentSymbolById, astNode.id, currentSymbols, ...fieldSymbols);
                }
            }
        }
    }
}
function addDocumentSymbols(nodeIdMapCollection, parentSymbolById, nodeId, currentSymbols, ...newSymbols) {
    const parentSymbol = findParentSymbol(nodeIdMapCollection, parentSymbolById, nodeId);
    if (parentSymbol) {
        if (!parentSymbol.children) {
            parentSymbol.children = [];
        }
        parentSymbol.children.push(...newSymbols);
        return;
    }
    // Add to the top level
    currentSymbols.push(...newSymbols);
}
function findParentSymbol(nodeIdMapCollection, parentSymbolById, nodeId) {
    // Get parent for current node
    const parentNodeId = nodeIdMapCollection.parentIdById.get(nodeId);
    if (!parentNodeId) {
        // No more parents to check
        return undefined;
    }
    let parentSymbol = parentSymbolById.get(parentNodeId);
    if (!parentSymbol) {
        parentSymbol = findParentSymbol(nodeIdMapCollection, parentSymbolById, parentNodeId);
    }
    return parentSymbol;
}
//# sourceMappingURL=documentSymbols.js.map