"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.tryAutocompleteFieldAccess = exports.inspectFieldAccess = void 0;
const PQP = require("@microsoft/powerquery-parser");
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const token_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language/token");
const trace_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/common/trace");
const activeNode_1 = require("../activeNode");
const __1 = require("../..");
const type_1 = require("../type");
function inspectFieldAccess(lexerSnapshot, nodeIdMapCollection, activeNode) {
    const fieldProjection = parser_1.AncestryUtils.findNodeKind(activeNode.ancestry, language_1.Ast.NodeKind.FieldProjection);
    const fieldSelector = parser_1.AncestryUtils.findNodeKind(activeNode.ancestry, language_1.Ast.NodeKind.FieldSelector);
    if (fieldProjection) {
        return inspectFieldProjection(lexerSnapshot, nodeIdMapCollection, activeNode.position, fieldProjection);
    }
    else if (fieldSelector) {
        return inspectFieldSelector(lexerSnapshot, nodeIdMapCollection, activeNode.position, fieldSelector);
    }
    else {
        return undefined;
    }
}
exports.inspectFieldAccess = inspectFieldAccess;
function tryAutocompleteFieldAccess(settings, parseState, activeNode, typeCache) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = settings.traceManager.entry(__1.AutocompleteTraceConstant.AutocompleteFieldAccess, tryAutocompleteFieldAccess.name, settings.initialCorrelationId);
        const updatedSettings = Object.assign(Object.assign({}, settings), { initialCorrelationId: trace.id });
        let result;
        if (!activeNode_1.ActiveNodeUtils.isPositionInBounds(activeNode)) {
            result = powerquery_parser_1.ResultUtils.ok(undefined);
        }
        else {
            result = yield powerquery_parser_1.ResultUtils.ensureResultAsync(() => autocompleteFieldAccess(updatedSettings, parseState, activeNode, typeCache), updatedSettings.locale);
        }
        trace.exit({ [trace_1.TraceConstant.IsError]: powerquery_parser_1.ResultUtils.isError(result) });
        return result;
    });
}
exports.tryAutocompleteFieldAccess = tryAutocompleteFieldAccess;
const AllowedExtendedTypeKindsForFieldEntries = [
    language_1.Type.ExtendedTypeKind.AnyUnion,
    language_1.Type.ExtendedTypeKind.DefinedRecord,
    language_1.Type.ExtendedTypeKind.DefinedTable,
];
function autocompleteFieldAccess(settings, parseState, activeNode, typeCache) {
    return __awaiter(this, void 0, void 0, function* () {
        const inspectedFieldAccess = inspectFieldAccess(parseState.lexerSnapshot, parseState.contextState.nodeIdMapCollection, activeNode);
        // No field access was found, or the field access reports no autocomplete is possible.
        // Eg. `[x = 1][x |]`
        if (!(inspectedFieldAccess === null || inspectedFieldAccess === void 0 ? void 0 : inspectedFieldAccess.isAutocompleteAllowed)) {
            return undefined;
        }
        // After a field access was found then find the field it's accessing and inspect the field's Type.
        // This is delayed until after the field access because running static type analysis on an
        // arbitrary field could be costly.
        const nodeIdMapCollection = parseState.contextState.nodeIdMapCollection;
        const field = typablePrimaryExpression(nodeIdMapCollection, activeNode);
        if (field === undefined) {
            return undefined;
        }
        const triedFieldType = yield (0, type_1.tryType)(settings, nodeIdMapCollection, field.node.id, typeCache);
        if (powerquery_parser_1.ResultUtils.isError(triedFieldType)) {
            throw triedFieldType.error;
        }
        const fieldType = triedFieldType.value;
        // We can only autocomplete a field access if we know what fields are present.
        const fieldEntries = fieldEntriesFromFieldType(fieldType);
        if (fieldEntries.length === 0) {
            return undefined;
        }
        return {
            field,
            fieldType,
            inspectedFieldAccess,
            autocompleteItems: createAutocompleteItems(fieldEntries, inspectedFieldAccess),
        };
    });
}
function createAutocompleteItems(fieldEntries, inspectedFieldAccess) {
    const autocompleteItems = [];
    const { textUnderPosition, textEditRange } = inspectedFieldAccess;
    for (const [label, powerQueryType] of fieldEntries) {
        autocompleteItems.push(createAutocompleteItem(label, powerQueryType, textUnderPosition, textEditRange));
    }
    return autocompleteItems;
}
function createAutocompleteItem(label, powerQueryType, textUnderPosition, textEditRange) {
    const jaroWinklerScore = textUnderPosition ? (0, __1.calculateJaroWinkler)(label, textUnderPosition) : 1;
    // If the key is a quoted identifier but doesn't need to be one then slice out the quote contents.
    const identifierKind = language_1.IdentifierUtils.getIdentifierKind(label, false);
    const normalizedLabel = identifierKind === language_1.IdentifierUtils.IdentifierKind.Quote ? label.slice(2, -1) : label;
    return {
        jaroWinklerScore,
        kind: __1.CompletionItemKind.Field,
        label: normalizedLabel,
        powerQueryType,
        textEdit: textEditRange ? __1.TextEdit.replace(textEditRange, label) : undefined,
    };
}
function emptyInspectedFieldAccess(isAutocompleteAllowed) {
    return {
        fieldNames: [],
        isAutocompleteAllowed,
        textEditRange: undefined,
        textUnderPosition: undefined,
    };
}
function fieldEntriesFromFieldType(type) {
    switch (type.extendedKind) {
        case language_1.Type.ExtendedTypeKind.AnyUnion: {
            let fields = [];
            for (const field of type.unionedTypePairs) {
                if (field.extendedKind && AllowedExtendedTypeKindsForFieldEntries.includes(field.extendedKind)) {
                    fields = fields.concat(fieldEntriesFromFieldType(field));
                }
            }
            return fields;
        }
        case language_1.Type.ExtendedTypeKind.DefinedRecord:
        case language_1.Type.ExtendedTypeKind.DefinedTable:
            return [...type.fields.entries()];
        default:
            return [];
    }
}
function inspectFieldProjection(lexerSnapshot, nodeIdMapCollection, position, fieldProjection) {
    const fieldNamesResult = [];
    let isAutocompleteAllowedResult = false;
    let textEditRangeResult;
    let textUnderPositionResult;
    for (const fieldSelector of parser_1.NodeIdMapIterator.iterFieldProjection(nodeIdMapCollection, fieldProjection)) {
        const inspectedFieldAccess = inspectFieldSelector(lexerSnapshot, nodeIdMapCollection, position, fieldSelector);
        fieldNamesResult.push(...inspectedFieldAccess.fieldNames);
        if (inspectedFieldAccess.isAutocompleteAllowed) {
            isAutocompleteAllowedResult = true;
            textEditRangeResult = inspectedFieldAccess.textEditRange;
            textUnderPositionResult = inspectedFieldAccess.textUnderPosition;
        }
    }
    return {
        fieldNames: fieldNamesResult,
        isAutocompleteAllowed: isAutocompleteAllowedResult,
        textEditRange: textEditRangeResult,
        textUnderPosition: textUnderPositionResult,
    };
}
function inspectFieldSelector(lexerSnapshot, nodeIdMapCollection, position, fieldSelector) {
    const childIds = nodeIdMapCollection.childIdsById.get(fieldSelector.node.id);
    if (childIds === undefined) {
        return emptyInspectedFieldAccess(false);
    }
    else if (childIds.length === 1) {
        return emptyInspectedFieldAccess(nodeIdMapCollection.astNodeById.has(childIds[0]));
    }
    const generalizedIdentifierXor = parser_1.NodeIdMapUtils.assertXorChecked(nodeIdMapCollection, PQP.ArrayUtils.assertGet(childIds, 1), language_1.Ast.NodeKind.GeneralizedIdentifier);
    switch (generalizedIdentifierXor.kind) {
        // There isn't a perfect way to determine if an autocomplete should be allowed.
        // Take the following scenario, staring with start with: `let foo = [key = 1] in foo`
        // Now the user wants to add a field selector: `let [key = 1][|in foo`
        // The parser rules generate the GeneralizedIdentifier `in foo`, even though the user wouldn't expect it.
        //
        // The best solution I can think of is the following:
        //  - if it' a fully parsed Ast (ie. there exists a closing bracket) then the entire field is fair game.
        //  - if it's a partial Ast (ie. there is no closing bracket) then only operate on the first identifier token
        case PQP.Parser.XorNodeKind.Ast: {
            const generalizedIdentifier = generalizedIdentifierXor.node;
            const generalizedIdentifierLiteral = generalizedIdentifier.literal;
            let isPositionInOrOnIdentifier;
            let textEditRange;
            let textUnderPosition;
            const firstToken = lexerSnapshot.tokens[generalizedIdentifier.tokenRange.tokenIndexStart];
            if (!generalizedIdentifierLiteral.includes(" ")) {
                isPositionInOrOnIdentifier = __1.PositionUtils.isInAst(position, generalizedIdentifier, true, true);
                textUnderPosition = isPositionInOrOnIdentifier ? generalizedIdentifierLiteral : undefined;
                textEditRange = __1.PositionUtils.rangeFromTokenRange(generalizedIdentifier.tokenRange);
            }
            // Given `let _ = [key with space = 1][#"key with space"| in _`
            // assume the user wants to autocomplete the identifier `#"key with space"`
            //
            // Given: `let _ = [key with space = 1][key| in _`
            // assume the user wants to autocomplete the identifier `key`
            else if ((firstToken === null || firstToken === void 0 ? void 0 : firstToken.kind) === token_1.TokenKind.Identifier) {
                isPositionInOrOnIdentifier = __1.PositionUtils.isInToken(position, firstToken, true, true);
                textUnderPosition = isPositionInOrOnIdentifier ? firstToken.data : undefined;
                textEditRange = isPositionInOrOnIdentifier ? __1.PositionUtils.rangeFromToken(firstToken) : undefined;
            }
            else {
                isPositionInOrOnIdentifier = false;
                textEditRange = undefined;
                textUnderPosition = undefined;
            }
            return {
                fieldNames: [generalizedIdentifierLiteral],
                isAutocompleteAllowed: isPositionInOrOnIdentifier,
                textEditRange,
                textUnderPosition,
            };
        }
        case PQP.Parser.XorNodeKind.Context: {
            return {
                fieldNames: [],
                isAutocompleteAllowed: true,
                textEditRange: undefined,
                textUnderPosition: undefined,
            };
        }
        default:
            throw powerquery_parser_1.Assert.isNever(generalizedIdentifierXor);
    }
}
function typablePrimaryExpression(nodeIdMapCollection, activeNode) {
    const ancestry = activeNode.ancestry;
    const numAncestors = ancestry.length;
    let contiguousPrimaryExpression;
    let matchingContiguousPrimaryExpression = true;
    for (let index = 0; index < numAncestors; index += 1) {
        const xorNode = ancestry[index];
        if (xorNode.node.kind === language_1.Ast.NodeKind.RecursivePrimaryExpression) {
            // The previous ancestor must be an attribute of Rpe, which is either its head or ArrrayWrapper.
            const xorNodeBeforeRpe = parser_1.AncestryUtils.assertNth(ancestry, index - 1);
            // If we're coming from the head node,
            // then return undefined as there can be no nodes before the head ode.
            if (xorNodeBeforeRpe.node.attributeIndex === 0) {
                return undefined;
            }
            // Else if we're coming from the ArrayWrapper,
            // then grab the previous sibling.
            else if (xorNodeBeforeRpe.node.attributeIndex === 1) {
                const rpeChild = parser_1.AncestryUtils.assertNth(ancestry, index - 2);
                return parser_1.NodeIdMapUtils.assertRecursiveExpressionPreviousSibling(nodeIdMapCollection, rpeChild.node.id);
            }
            else {
                throw new PQP.CommonError.InvariantError(`the child of a ${language_1.Ast.NodeKind.RecursivePrimaryExpression} should have an attribute index of either 1 or 2`, {
                    parentId: xorNode.node.id,
                    childId: xorNodeBeforeRpe.node.id,
                });
            }
        }
        else if (matchingContiguousPrimaryExpression && parser_1.XorNodeUtils.isTPrimaryExpression(xorNode)) {
            contiguousPrimaryExpression = xorNode;
        }
        else {
            matchingContiguousPrimaryExpression = false;
        }
    }
    return contiguousPrimaryExpression;
}
//# sourceMappingURL=autocompleteFieldAccess.js.map