"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.tryAutocompleteLanguageConstant = void 0;
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const trace_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/common/trace");
const constant_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language/constant/constant");
const activeNode_1 = require("../activeNode");
const __1 = require("../..");
function tryAutocompleteLanguageConstant(settings, nodeIdMapCollection, activeNode, trailingToken) {
    const trace = settings.traceManager.entry(__1.AutocompleteTraceConstant.AutocompleteLanguageConstant, tryAutocompleteLanguageConstant.name, settings.initialCorrelationId);
    const result = powerquery_parser_1.ResultUtils.ensureResult(() => autocompleteLanguageConstant(nodeIdMapCollection, activeNode, trailingToken), settings.locale);
    trace.exit({ [trace_1.TraceConstant.IsError]: powerquery_parser_1.ResultUtils.isError(result) });
    return result;
}
exports.tryAutocompleteLanguageConstant = tryAutocompleteLanguageConstant;
function autocompleteLanguageConstant(nodeIdMapCollection, activeNode, trailingToken) {
    if (!activeNode_1.ActiveNodeUtils.isPositionInBounds(activeNode)) {
        return [];
    }
    return [
        getCatchAutocompleteItem(nodeIdMapCollection, activeNode, trailingToken),
        getNullableAutocompleteItem(activeNode, trailingToken),
        getOptionalAutocompleteItem(activeNode),
    ].filter(powerquery_parser_1.TypeScriptUtils.isDefined);
}
function createAutocompleteItem(label, other, range) {
    const jaroWinklerScore = other !== undefined ? (0, __1.calculateJaroWinkler)(label, other) : 1;
    return {
        jaroWinklerScore,
        kind: __1.CompletionItemKind.Keyword,
        label,
        powerQueryType: language_1.Type.NotApplicableInstance,
        textEdit: range ? vscode_languageserver_types_1.TextEdit.replace(range, label) : undefined,
    };
}
function getCatchAutocompleteItem(nodeIdMapCollection, activeNode, trailingToken) {
    const nodeIndex = parser_1.AncestryUtils.indexOfNodeKind(activeNode.ancestry, language_1.Ast.NodeKind.ErrorHandlingExpression);
    if (nodeIndex === undefined) {
        return undefined;
    }
    const errorHandlingExpression = parser_1.AncestryUtils.nthChecked(activeNode.ancestry, nodeIndex, language_1.Ast.NodeKind.ErrorHandlingExpression);
    if (!errorHandlingExpression) {
        return undefined;
    }
    switch (errorHandlingExpression.kind) {
        case parser_1.XorNodeKind.Ast:
            // `try 1 catch| (exc) => 1`
            // `try 1 otherwise| 1
            if (errorHandlingExpression.node.handler) {
                const handler = errorHandlingExpression.node.handler;
                if (handler.constant && __1.PositionUtils.isInAst(activeNode.position, handler.constant, true, true)) {
                    return createAutocompleteItem(constant_1.LanguageConstant.Catch, handler.constant.constantKind, __1.PositionUtils.rangeFromTokenRange(handler.constant.tokenRange));
                }
            }
            else if (__1.PositionUtils.isAfterAst(activeNode.position, errorHandlingExpression.node, true)) {
                // `try 1 |`
                if (!trailingToken) {
                    return createAutocompleteItem(constant_1.LanguageConstant.Catch);
                }
                // `try 1 cat|`
                else if (__1.PositionUtils.isInToken(activeNode.position, trailingToken, true, true)) {
                    return createAutocompleteItem(constant_1.LanguageConstant.Catch, trailingToken.data, __1.PositionUtils.rangeFromToken(trailingToken));
                }
            }
            break;
        case parser_1.XorNodeKind.Context: {
            const handler = parser_1.AncestryUtils.nthChecked(activeNode.ancestry, nodeIndex - 1, [language_1.Ast.NodeKind.CatchExpression, language_1.Ast.NodeKind.OtherwiseExpression]);
            if (handler === undefined) {
                return undefined;
            }
            const catchConstant = parser_1.NodeIdMapUtils.nthChildAstChecked(nodeIdMapCollection, handler.node.id, 0, language_1.Ast.NodeKind.Constant);
            if (catchConstant && __1.PositionUtils.isInAst(activeNode.position, catchConstant, true, true)) {
                return createAutocompleteItem(constant_1.LanguageConstant.Catch, catchConstant.constantKind, __1.PositionUtils.rangeFromTokenRange(catchConstant.tokenRange));
            }
            break;
        }
        default:
            powerquery_parser_1.Assert.isNever(errorHandlingExpression);
    }
    return undefined;
}
function getNullableAutocompleteItem(activeNode, trailingToken) {
    const ancestry = activeNode.ancestry;
    const numAncestors = ancestry.length;
    for (let index = 0; index < numAncestors; index += 1) {
        const xorNode = ancestry[index];
        switch (xorNode.node.kind) {
            case language_1.Ast.NodeKind.AsNullablePrimitiveType:
                if (isNullableAllowedForAsNullablePrimitiveType(activeNode, index)) {
                    return createAutocompleteItem(constant_1.LanguageConstant.Nullable);
                }
                break;
            case language_1.Ast.NodeKind.PrimitiveType:
                if (parser_1.XorNodeUtils.isContext(xorNode)) {
                    if (!trailingToken) {
                        return createAutocompleteItem(constant_1.LanguageConstant.Nullable);
                    }
                    else if (__1.PositionUtils.isInToken(activeNode.position, trailingToken, true, true)) {
                        return createAutocompleteItem(constant_1.LanguageConstant.Nullable, trailingToken.data, __1.PositionUtils.rangeFromToken(trailingToken));
                    }
                }
                break;
            default:
                continue;
        }
    }
    return undefined;
}
function getOptionalAutocompleteItem(activeNode) {
    const fnExprAncestryIndex = parser_1.AncestryUtils.indexOfNodeKind(activeNode.ancestry, language_1.Ast.NodeKind.FunctionExpression);
    if (fnExprAncestryIndex === undefined) {
        return undefined;
    }
    // FunctionExpression -> IParenthesisWrapped -> ParameterList -> Csv -> Parameter
    const parameterAncestryIndex = fnExprAncestryIndex - 4;
    const parameter = parser_1.AncestryUtils.nthChecked(activeNode.ancestry, parameterAncestryIndex, language_1.Ast.NodeKind.Parameter);
    if (parameter === undefined) {
        return undefined;
    }
    const childOfParameter = parser_1.AncestryUtils.nth(activeNode.ancestry, parameterAncestryIndex - 1);
    if (childOfParameter === undefined) {
        return createAutocompleteItem(constant_1.LanguageConstant.Optional);
    }
    switch (childOfParameter.node.attributeIndex) {
        // IParameter.optionalConstant
        case 0:
            switch (childOfParameter.kind) {
                case parser_1.XorNodeKind.Ast: {
                    const optionalConstant = parser_1.XorNodeUtils.assertAstChecked(childOfParameter, language_1.Ast.NodeKind.Constant);
                    return createAutocompleteItem(constant_1.LanguageConstant.Optional, optionalConstant.constantKind, __1.PositionUtils.rangeFromTokenRange(optionalConstant.tokenRange));
                }
                case parser_1.XorNodeKind.Context:
                    return createAutocompleteItem(constant_1.LanguageConstant.Optional);
                default:
                    throw powerquery_parser_1.Assert.isNever(childOfParameter);
            }
        // IParameter.name
        case 1:
            switch (childOfParameter.kind) {
                case parser_1.XorNodeKind.Ast: {
                    const parameterName = parser_1.XorNodeUtils.assertAstChecked(childOfParameter, language_1.Ast.NodeKind.Identifier);
                    return createAutocompleteItem(constant_1.LanguageConstant.Optional, parameterName.literal, __1.PositionUtils.rangeFromTokenRange(parameterName.tokenRange));
                }
                case parser_1.XorNodeKind.Context:
                    return createAutocompleteItem(constant_1.LanguageConstant.Optional);
                default:
                    throw powerquery_parser_1.Assert.isNever(childOfParameter);
            }
        default:
            return undefined;
    }
}
function isNullableAllowedForAsNullablePrimitiveType(activeNode, ancestryIndex) {
    const child = parser_1.AncestryUtils.nth(activeNode.ancestry, ancestryIndex - 1);
    if ((child === null || child === void 0 ? void 0 : child.node.attributeIndex) !== 1) {
        return false;
    }
    // Ast.AsNullablePrimitiveType.paired: Ast.TNullablePrimitiveType
    const paired = child;
    const position = activeNode.position;
    // Ast.PrimitiveType
    if (paired.node.kind === language_1.Ast.NodeKind.PrimitiveType && __1.PositionUtils.isBeforeXor(position, paired, false)) {
        return true;
    }
    // Ast.NullablePrimitiveType
    else if (paired.node.kind === language_1.Ast.NodeKind.NullablePrimitiveType) {
        const grandchild = parser_1.AncestryUtils.nth(activeNode.ancestry, ancestryIndex - 2);
        if (grandchild === undefined) {
            return false;
        }
        return (
        // Ast.Constant
        grandchild.node.kind === language_1.Ast.NodeKind.Constant ||
            // before Ast.PrimitiveType
            __1.PositionUtils.isBeforeXor(position, grandchild, false));
    }
    else if (paired.node.kind === language_1.Ast.NodeKind.PrimitiveType) {
        return parser_1.XorNodeUtils.isContext(paired);
    }
    else {
        return false;
    }
}
//# sourceMappingURL=autocompleteLanguageConstant.js.map