"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.expectedType = exports.findExpectedType = exports.tryExpectedType = void 0;
const PQP = require("@microsoft/powerquery-parser");
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const activeNode_1 = require("./activeNode");
const trace_1 = require("../trace");
function tryExpectedType(settings, activeNode) {
    const trace = settings.traceManager.entry(trace_1.InspectionTraceConstant.InspectExpectedType, tryExpectedType.name, settings.initialCorrelationId);
    if (!activeNode_1.ActiveNodeUtils.isPositionInBounds(activeNode)) {
        trace.exit();
        return powerquery_parser_1.ResultUtils.ok(undefined);
    }
    const result = powerquery_parser_1.ResultUtils.ensureResult(() => findExpectedType(activeNode), settings.locale);
    trace.exit();
    return result;
}
exports.tryExpectedType = tryExpectedType;
// Traverse up the ancestry and find what type is expected as the nth child of a node's kind.
// The last type generated this way should have the widest typing,
// which then can be used for type hinting.
function findExpectedType(activeNode) {
    const ancestry = activeNode.ancestry;
    const upperBound = ancestry.length - 1;
    let bestMatch;
    for (let index = 0; index < upperBound; index += 1) {
        const parent = ancestry[index + 1];
        const child = ancestry[index];
        const childAttributeIndex = powerquery_parser_1.Assert.asDefined(child.node.attributeIndex, `Expected child to have an attribute index.`, { childId: child.node.id });
        const attributeIndex = parser_1.XorNodeUtils.isAst(parent) && activeNode.leafKind === activeNode_1.ActiveNodeLeafKind.IsAfterPosition
            ? childAttributeIndex + 1
            : childAttributeIndex;
        const allowedType = expectedType(parent, attributeIndex);
        if (allowedType.kind !== language_1.Type.TypeKind.NotApplicable) {
            bestMatch = allowedType;
        }
    }
    return bestMatch;
}
exports.findExpectedType = findExpectedType;
// For a given parent node, what is the expected type for a child at a given index?
function expectedType(parentXorNode, childIndex) {
    switch (parentXorNode.node.kind) {
        case language_1.Ast.NodeKind.ArrayWrapper:
        case language_1.Ast.NodeKind.Constant:
        case language_1.Ast.NodeKind.Csv:
        case language_1.Ast.NodeKind.GeneralizedIdentifier:
        case language_1.Ast.NodeKind.Identifier:
        case language_1.Ast.NodeKind.IdentifierExpression:
        case language_1.Ast.NodeKind.InvokeExpression:
        case language_1.Ast.NodeKind.FieldProjection:
        case language_1.Ast.NodeKind.FieldSelector:
        case language_1.Ast.NodeKind.FieldSpecificationList:
        case language_1.Ast.NodeKind.ListExpression:
        case language_1.Ast.NodeKind.ListLiteral:
        case language_1.Ast.NodeKind.LiteralExpression:
        case language_1.Ast.NodeKind.RecordLiteral:
        case language_1.Ast.NodeKind.RecordType:
        case language_1.Ast.NodeKind.Parameter:
        case language_1.Ast.NodeKind.ParameterList:
        case language_1.Ast.NodeKind.PrimitiveType:
        case language_1.Ast.NodeKind.RecordExpression:
        case language_1.Ast.NodeKind.RecursivePrimaryExpression:
            return language_1.Type.NotApplicableInstance;
        case language_1.Ast.NodeKind.ArithmeticExpression:
        case language_1.Ast.NodeKind.EqualityExpression:
        case language_1.Ast.NodeKind.LogicalExpression:
        case language_1.Ast.NodeKind.RelationalExpression:
            return expectTypeBinOpExpression(parentXorNode, childIndex);
        case language_1.Ast.NodeKind.AsExpression:
        case language_1.Ast.NodeKind.IsExpression:
            return expectTypeAsOrIsExpression(parentXorNode, childIndex);
        case language_1.Ast.NodeKind.CatchExpression:
        case language_1.Ast.NodeKind.OtherwiseExpression:
            return expectTypeErrorHandler(parentXorNode, childIndex);
        case language_1.Ast.NodeKind.GeneralizedIdentifierPairedExpression:
        case language_1.Ast.NodeKind.IdentifierPairedExpression:
            return expectTypeIdentifierPairedExpression(parentXorNode, childIndex);
        case language_1.Ast.NodeKind.Section:
        case language_1.Ast.NodeKind.SectionMember:
            return expectTypeSectionOrSectionMember(childIndex);
        case language_1.Ast.NodeKind.AsType:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.TypeProductionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.AsNullablePrimitiveType:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.NullablePrimitiveInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.EachExpression:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.ExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.ErrorHandlingExpression:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.ExpressionInstance;
                case 2:
                    return language_1.Type.NotApplicableInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        // TODO: how should error raising be typed?
        case language_1.Ast.NodeKind.ErrorRaisingExpression:
            return language_1.Type.NotApplicableInstance;
        case language_1.Ast.NodeKind.GeneralizedIdentifierPairedAnyLiteral:
            switch (childIndex) {
                case 0:
                case 1:
                    return language_1.Type.NotApplicableInstance;
                case 2:
                    return language_1.Type.AnyLiteralInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.FieldSpecification:
            switch (childIndex) {
                case 0:
                case 1:
                    return language_1.Type.NotApplicableInstance;
                case 2:
                    return language_1.Type.TypeProductionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.FieldTypeSpecification:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.TypeProductionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.FunctionExpression:
            switch (childIndex) {
                case 0:
                case 2:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.NullablePrimitiveInstance;
                case 3:
                    return language_1.Type.ExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.FunctionType:
            switch (childIndex) {
                case 0:
                case 1:
                    return language_1.Type.NotApplicableInstance;
                case 2:
                    return language_1.Type.NullablePrimitiveInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.IfExpression:
            switch (childIndex) {
                case 0:
                case 2:
                case 4:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.LogicalInstance;
                case 3:
                case 5:
                    return language_1.Type.ExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.IsNullablePrimitiveType:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.NullablePrimitiveInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.ItemAccessExpression:
            switch (childIndex) {
                case 0:
                case 2:
                case 3:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.ExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.LetExpression:
            switch (childIndex) {
                case 0:
                case 1:
                case 2:
                    return language_1.Type.NotApplicableInstance;
                case 3:
                    return language_1.Type.ExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.ListType:
            switch (childIndex) {
                case 0:
                case 2:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.TypeProductionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.MetadataExpression:
            switch (childIndex) {
                case 1:
                    return language_1.Type.NotApplicableInstance;
                case 0:
                case 2:
                    return language_1.Type.TypeExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.NotImplementedExpression:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.NullCoalescingExpression:
            switch (childIndex) {
                case 0:
                case 2:
                    return language_1.Type.ExpressionInstance;
                case 1:
                    return language_1.Type.NotApplicableInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.NullablePrimitiveType:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.PrimitiveInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.NullableType:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.TypeProductionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.ParenthesizedExpression:
            switch (childIndex) {
                case 0:
                case 2:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.ExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.RangeExpression:
            switch (childIndex) {
                case 0:
                case 2:
                    return language_1.Type.ExpressionInstance;
                case 1:
                    return language_1.Type.NotApplicableInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.TableType:
            switch (childIndex) {
                case 1:
                case 2:
                    return language_1.Type.PrimaryExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.TypePrimaryType:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.PrimaryTypeInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        case language_1.Ast.NodeKind.UnaryExpression:
            switch (childIndex) {
                case 0:
                    return language_1.Type.NotApplicableInstance;
                case 1:
                    return language_1.Type.TypeExpressionInstance;
                default:
                    throw unknownChildIndexError(parentXorNode, childIndex);
            }
        default:
            throw powerquery_parser_1.Assert.isNever(parentXorNode.node);
    }
}
exports.expectedType = expectedType;
function expectTypeAsOrIsExpression(parentXorNode, childIndex) {
    switch (childIndex) {
        case 0:
            return language_1.Type.TypeExpressionInstance;
        case 1:
            return language_1.Type.NotApplicableInstance;
        case 2:
            return language_1.Type.NullablePrimitiveInstance;
        default:
            throw unknownChildIndexError(parentXorNode, childIndex);
    }
}
function expectTypeErrorHandler(parentXorNode, childIndex) {
    switch (childIndex) {
        case 0:
            return language_1.Type.NotApplicableInstance;
        case 1:
            return language_1.Type.ExpressionInstance;
        default:
            throw unknownChildIndexError(parentXorNode, childIndex);
    }
}
function expectTypeBinOpExpression(parentXorNode, childIndex) {
    switch (childIndex) {
        case 0:
        case 2:
            return language_1.Type.TypeExpressionInstance;
        case 1:
            return language_1.Type.NotApplicableInstance;
        default:
            throw unknownChildIndexError(parentXorNode, childIndex);
    }
}
function expectTypeIdentifierPairedExpression(parentXorNode, childIndex) {
    switch (childIndex) {
        case 0:
        case 1:
            return language_1.Type.NotApplicableInstance;
        case 2:
            return language_1.Type.ExpressionInstance;
        default:
            throw unknownChildIndexError(parentXorNode, childIndex);
    }
}
function expectTypeSectionOrSectionMember(childIndex) {
    switch (childIndex) {
        case 0:
            return language_1.Type.RecordInstance;
        default:
            return language_1.Type.NotApplicableInstance;
    }
}
function unknownChildIndexError(parent, childIndex) {
    const details = {
        parentId: parent.node.kind,
        parentNodeKind: parent.node.kind,
        childIndex,
    };
    return new PQP.CommonError.InvariantError(`unknown childIndex`, details);
}
//# sourceMappingURL=expectedType.js.map