"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.tryInvokeExpression = void 0;
const PQP = require("@microsoft/powerquery-parser");
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const trace_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/common/trace");
const scope_1 = require("../scope");
const type_1 = require("../type");
const typeCache_1 = require("../typeCache");
const trace_2 = require("../../trace");
function tryInvokeExpression(settings, nodeIdMapCollection, invokeExpressionId, 
// If a TypeCache is given, then potentially add to its values and include it as part of the return,
// Else create a new TypeCache and include it in the return.
typeCache = typeCache_1.TypeCacheUtils.emptyCache()) {
    const trace = settings.traceManager.entry(trace_2.InspectionTraceConstant.InspectInvokeExpression, tryInvokeExpression.name, settings.initialCorrelationId);
    const updatedSettings = Object.assign(Object.assign({}, settings), { initialCorrelationId: trace.id });
    const result = powerquery_parser_1.ResultUtils.ensureResultAsync(() => inspectInvokeExpression(updatedSettings, nodeIdMapCollection, invokeExpressionId, typeCache, trace.id), settings.locale);
    return result;
}
exports.tryInvokeExpression = tryInvokeExpression;
function inspectInvokeExpression(settings, nodeIdMapCollection, invokeExpressionId, typeCache, correlationId) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const trace = settings.traceManager.entry(trace_2.InspectionTraceConstant.InspectInvokeExpression, inspectInvokeExpression.name, correlationId);
        (_a = settings.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
        const invokeExpressionXorNode = parser_1.NodeIdMapUtils.xor(nodeIdMapCollection, invokeExpressionId);
        if (invokeExpressionXorNode === undefined) {
            trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
            throw new PQP.CommonError.InvariantError(`expected invokeExpressionId to be present in nodeIdMapCollection`, {
                invokeExpressionId,
            });
        }
        parser_1.XorNodeUtils.assertIsNodeKind(invokeExpressionXorNode, language_1.Ast.NodeKind.InvokeExpression);
        const previousNode = parser_1.NodeIdMapUtils.assertRecursiveExpressionPreviousSibling(nodeIdMapCollection, invokeExpressionId);
        const functionType = powerquery_parser_1.ResultUtils.assertOk(yield (0, type_1.tryType)(settings, nodeIdMapCollection, previousNode.node.id, typeCache));
        const name = parser_1.NodeIdMapUtils.invokeExpressionIdentifierLiteral(nodeIdMapCollection, invokeExpressionId);
        let invokeExpressionArgs;
        if (language_1.TypeUtils.isDefinedFunction(functionType)) {
            const iterableArguments = parser_1.NodeIdMapIterator.iterInvokeExpression(nodeIdMapCollection, invokeExpressionXorNode);
            const givenArgumentTypes = yield getArgumentTypes(settings, nodeIdMapCollection, typeCache, iterableArguments);
            const givenArguments = iterableArguments.slice(0, givenArgumentTypes.length);
            const [numMinExpectedArguments, numMaxExpectedArguments] = getNumExpectedArguments(functionType);
            invokeExpressionArgs = {
                givenArguments,
                givenArgumentTypes,
                numMaxExpectedArguments,
                numMinExpectedArguments,
                typeChecked: language_1.TypeUtils.typeCheckInvocation(givenArgumentTypes, functionType, settings.traceManager, trace.id),
            };
        }
        const result = {
            invokeExpressionXorNode,
            functionType,
            isNameInLocalScope: yield getIsNameInLocalScope(settings, nodeIdMapCollection, typeCache, invokeExpressionXorNode, name, trace.id),
            name,
            arguments: invokeExpressionArgs,
        };
        trace.exit();
        return result;
    });
}
function getIsNameInLocalScope(settings, nodeIdMapCollection, typeCache, invokeExpressionXorNode, name, correlationId) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = settings.traceManager.entry(trace_2.InspectionTraceConstant.InspectInvokeExpression, getIsNameInLocalScope.name, correlationId);
        const updatedSettings = Object.assign(Object.assign({}, settings), { initialCorrelationId: trace.id });
        // Try to find out if the identifier is a local or external name.
        if (name !== undefined) {
            // Seed local scope
            const scope = powerquery_parser_1.ResultUtils.assertOk(yield (0, scope_1.assertGetOrCreateNodeScope)(updatedSettings, nodeIdMapCollection, updatedSettings.eachScopeById, invokeExpressionXorNode.node.id, typeCache.scopeById));
            const nameScopeItem = scope.get(name);
            trace.exit();
            return nameScopeItem !== undefined && nameScopeItem.kind !== scope_1.ScopeItemKind.Undefined;
        }
        else {
            trace.exit();
            return false;
        }
    });
}
function getNumExpectedArguments(functionType) {
    const nonOptionalArguments = functionType.parameters.filter((parameter) => !parameter.isOptional);
    const numMinExpectedArguments = nonOptionalArguments.length;
    const numMaxExpectedArguments = functionType.parameters.length;
    return [numMinExpectedArguments, numMaxExpectedArguments];
}
function getArgumentTypes(settings, nodeIdMapCollection, typeCache, argXorNodes) {
    return __awaiter(this, void 0, void 0, function* () {
        const result = [];
        for (const xorNode of argXorNodes) {
            // eslint-disable-next-line no-await-in-loop
            const triedArgType = yield (0, type_1.tryType)(settings, nodeIdMapCollection, xorNode.node.id, typeCache);
            if (powerquery_parser_1.ResultUtils.isError(triedArgType)) {
                throw triedArgType;
            }
            const argType = triedArgType.value;
            // Occurs when there's expected to be a trailing argument, but none exist.
            // Eg. `foo(|` will iterate over an TXorNode which: contains no parsed elements, and evaluates to unknown.
            if (language_1.TypeUtils.isUnknown(argType) && !parser_1.NodeIdMapUtils.hasParsedToken(nodeIdMapCollection, xorNode.node.id)) {
                powerquery_parser_1.Assert.isTrue(xorNode === argXorNodes[argXorNodes.length - 1]);
                return result;
            }
            result.push(argType);
        }
        return result;
    });
}
//# sourceMappingURL=invokeExpression.js.map