"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.calculateJaroWinklers = exports.calculateJaroWinkler = exports.calculateJaro = void 0;
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
// https://en.wikipedia.org/wiki/Jaro%E2%80%93Winkler_distance
// The standard magic numbers for Winkler adjustments.
const WinklerAdjustment = 0.1;
const WinklerMaxCharacterAdjustment = 4;
const WinklerScoreThreshold = 0.7;
function calculateJaro(left, right) {
    const [shorter, longer] = getNormalized(left, right);
    return calculateJaroForNormalized(shorter, longer);
}
exports.calculateJaro = calculateJaro;
function calculateJaroWinkler(left, right) {
    const [shorter, longer] = getNormalized(left, right);
    let jaroScore = calculateJaroForNormalized(shorter, longer);
    if (jaroScore <= WinklerScoreThreshold) {
        return jaroScore;
    }
    // Adjust score for the number of leading characters that match.
    // Capped at an upper bound of N leading characters, where N is most commonly `4`.
    const upperBound = Math.max(WinklerMaxCharacterAdjustment, shorter.length);
    let startingCharIndex = 0;
    while (shorter[startingCharIndex] === longer[startingCharIndex] && startingCharIndex < upperBound) {
        startingCharIndex += 1;
    }
    if (startingCharIndex) {
        jaroScore += startingCharIndex * WinklerAdjustment * (1 - jaroScore);
    }
    return jaroScore;
}
exports.calculateJaroWinkler = calculateJaroWinkler;
// Returns [score, string] for the string with the highest jaro winkler score.
function calculateJaroWinklers(source, options) {
    powerquery_parser_1.ArrayUtils.assertNonZeroLength(options);
    let bestScore = -1;
    let bestOption = "";
    for (const option of options) {
        const currentScore = calculateJaroWinkler(source, option);
        if (currentScore > bestScore) {
            bestScore = currentScore;
            bestOption = option;
        }
    }
    return [bestScore, bestOption];
}
exports.calculateJaroWinklers = calculateJaroWinklers;
function calculateJaroForNormalized(shorter, longer) {
    const shorterLength = shorter.length;
    const longerLength = longer.length;
    // When trying to find similar characters limit the search to a window based on longer's length.
    const matchingWindowSize = Math.floor(longerLength / 2) - 1;
    const shorterMatches = new Array(shorterLength);
    const longerMatches = new Array(longerLength);
    let numMatches = 0;
    for (let longerIndex = 0; longerIndex < longerLength; longerIndex += 1) {
        const longerChar = longer[longerIndex];
        const matchingWindowStart = Math.max(0, longerIndex - matchingWindowSize);
        const matchingWindowEnd = Math.min(longerIndex + matchingWindowSize + 1, shorterLength);
        for (let shorterIndex = matchingWindowStart; shorterIndex < matchingWindowEnd; shorterIndex += 1) {
            const shorterChar = shorter[shorterIndex];
            if (longerChar === shorterChar && !shorterMatches[shorterIndex]) {
                shorterMatches[shorterIndex] = true;
                longerMatches[longerIndex] = true;
                numMatches += 1;
                break;
            }
        }
    }
    if (!numMatches) {
        return 0;
    }
    // Next we calculate the number of transpositions needed to convert `left` into `right` (or vice-versa).
    // `k` is part of an optimization.
    // It's used to keep a persistent index as we iterate over longerLength,
    // which will prevent double matching or a double for-loop.
    let k = 0;
    let numTranspositionsNeeded = 0;
    for (let longerIndex = 0; longerIndex < longerLength; longerIndex += 1) {
        if (!longerMatches[longerIndex]) {
            continue;
        }
        while (!shorterMatches[k]) {
            k += 1;
        }
        if (longer[longerIndex] !== shorter[k]) {
            numTranspositionsNeeded += 1;
        }
        k += 1;
    }
    const m_over_s1 = numMatches / longerLength;
    const m_over_s2 = numMatches / shorterLength;
    const mMinusT_over_m = (numMatches - numTranspositionsNeeded) / numMatches;
    return (m_over_s1 + m_over_s2 + mMinusT_over_m) / 3;
}
function getNormalized(left, right) {
    left = left.toLowerCase();
    right = right.toLowerCase();
    return left.length < right.length ? [left, right] : [right, left];
}
//# sourceMappingURL=jaroWinkler.js.map