"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.createLibraryDefinitions = exports.createLibraryDefinition = exports.createLibrary = exports.FailedLibrarySymbolConversionKind = void 0;
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const trace_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/common/trace");
const externalType_1 = require("../externalType");
const library_1 = require("../library");
const commonTypes_1 = require("../commonTypes");
var FailedLibrarySymbolConversionKind;
(function (FailedLibrarySymbolConversionKind) {
    FailedLibrarySymbolConversionKind["CompletionItemKind"] = "CompletionItemKind";
    FailedLibrarySymbolConversionKind["DefinedFunction"] = "DefinedFunction";
    FailedLibrarySymbolConversionKind["Parameter"] = "Parameter";
    FailedLibrarySymbolConversionKind["PrimitiveType"] = "PrimitiveType";
})(FailedLibrarySymbolConversionKind = exports.FailedLibrarySymbolConversionKind || (exports.FailedLibrarySymbolConversionKind = {}));
/**
    Always returns a library, even if there are failed library symbol conversions.
    An Ok result indicates that all library symbols were successfully converted.
    An Error result indicates that some library symbols failed to convert.
*/
function createLibrary(librarySymbols, dynamicLibraryDefinitions, externalTypeResolverFn) {
    const libraryDefinitionsResult = createLibraryDefinitions(librarySymbols);
    let staticLibraryDefinitions;
    let failedLibrarySymbolConversions;
    if (powerquery_parser_1.ResultUtils.isOk(libraryDefinitionsResult)) {
        staticLibraryDefinitions = libraryDefinitionsResult.value;
        failedLibrarySymbolConversions = [];
    }
    else {
        staticLibraryDefinitions = libraryDefinitionsResult.error.libraryDefinitions;
        failedLibrarySymbolConversions = libraryDefinitionsResult.error.failedLibrarySymbolConversions;
    }
    const libraryDefinitions = {
        dynamicLibraryDefinitions,
        staticLibraryDefinitions,
    };
    const definitionResolverFn = library_1.LibraryDefinitionUtils.externalTypeResolver(libraryDefinitions);
    const library = {
        externalTypeResolver: externalTypeResolverFn
            ? externalType_1.ExternalTypeUtils.composeExternalTypeResolvers(externalTypeResolverFn, definitionResolverFn)
            : definitionResolverFn,
        libraryDefinitions,
    };
    if (failedLibrarySymbolConversions.length > 0) {
        return powerquery_parser_1.ResultUtils.error({
            library,
            failedLibrarySymbolConversions,
        });
    }
    else {
        return powerquery_parser_1.ResultUtils.ok(library);
    }
}
exports.createLibrary = createLibrary;
function createLibraryDefinition(librarySymbol) {
    var _a, _b;
    const primitiveType = stringToPrimitiveType(librarySymbol.type);
    const completionItemKind = numberToCompletionItemKind(librarySymbol.completionItemKind);
    if (primitiveType === undefined) {
        return failedConversionError(librarySymbol, FailedLibrarySymbolConversionKind.PrimitiveType);
    }
    if (completionItemKind === undefined) {
        return failedConversionError(librarySymbol, FailedLibrarySymbolConversionKind.CompletionItemKind);
    }
    const label = librarySymbol.name;
    const description = (_b = (_a = librarySymbol.documentation) === null || _a === void 0 ? void 0 : _a.description) !== null && _b !== void 0 ? _b : "No description available";
    if (primitiveType.kind === language_1.Type.TypeKind.Type) {
        return powerquery_parser_1.ResultUtils.ok({
            kind: library_1.Library.LibraryDefinitionKind.Type,
            label,
            description,
            asPowerQueryType: primitiveType,
            completionItemKind,
        });
    }
    else if (librarySymbol.functionParameters) {
        const definedFunction = librarySymbolFunctionParamatersToDefinedFunction(librarySymbol, primitiveType);
        if (definedFunction === undefined) {
            return failedConversionError(librarySymbol, FailedLibrarySymbolConversionKind.DefinedFunction);
        }
        const parameters = [];
        for (const [parameter, index] of powerquery_parser_1.ArrayUtils.enumerate(librarySymbol.functionParameters)) {
            const libraryParameter = librarySymbolFunctionParameterToLibraryParameter(parameter);
            if (libraryParameter === undefined) {
                return failedParameterConversionError(librarySymbol, index);
            }
            parameters.push(libraryParameter);
        }
        return powerquery_parser_1.ResultUtils.ok({
            kind: library_1.Library.LibraryDefinitionKind.Function,
            label: language_1.TypeUtils.nameOf(definedFunction, trace_1.NoOpTraceManagerInstance, undefined),
            description,
            asPowerQueryType: definedFunction,
            completionItemKind,
            parameters,
        });
    }
    else {
        return powerquery_parser_1.ResultUtils.ok({
            kind: library_1.Library.LibraryDefinitionKind.Constant,
            label,
            description,
            asPowerQueryType: primitiveType,
            completionItemKind,
        });
    }
}
exports.createLibraryDefinition = createLibraryDefinition;
function createLibraryDefinitions(librarySymbols) {
    const libraryDefinitions = new Map();
    const failedLibrarySymbolConversions = [];
    for (const librarySymbol of librarySymbols) {
        const libraryDefinitionResult = createLibraryDefinition(librarySymbol);
        if (powerquery_parser_1.ResultUtils.isOk(libraryDefinitionResult)) {
            libraryDefinitions.set(librarySymbol.name, libraryDefinitionResult.value);
        }
        else {
            failedLibrarySymbolConversions.push(libraryDefinitionResult.error);
        }
    }
    if (failedLibrarySymbolConversions.length === 0) {
        return powerquery_parser_1.ResultUtils.ok(libraryDefinitions);
    }
    else {
        return powerquery_parser_1.ResultUtils.error({
            libraryDefinitions,
            failedLibrarySymbolConversions,
        });
    }
}
exports.createLibraryDefinitions = createLibraryDefinitions;
function failedConversionError(librarySymbol, kind) {
    return powerquery_parser_1.ResultUtils.error({
        librarySymbol,
        kind,
    });
}
function failedParameterConversionError(librarySymbol, index) {
    return powerquery_parser_1.ResultUtils.error({
        librarySymbol,
        kind: FailedLibrarySymbolConversionKind.Parameter,
        index,
    });
}
function librarySymbolFunctionParamatersToDefinedFunction(librarySymbol, returnType) {
    var _a;
    const parameters = [];
    for (const parameter of (_a = librarySymbol.functionParameters) !== null && _a !== void 0 ? _a : []) {
        const primitiveType = stringToPrimitiveType(parameter.type);
        if (primitiveType === undefined) {
            return undefined;
        }
        parameters.push({
            isNullable: primitiveType.isNullable,
            isOptional: !parameter.isRequired,
            type: primitiveType.kind,
            nameLiteral: parameter.name,
        });
    }
    return language_1.TypeUtils.definedFunction(false, parameters, returnType);
}
function librarySymbolFunctionParameterToLibraryParameter(parameter) {
    var _a;
    const primitiveType = stringToPrimitiveType(parameter.type);
    if (primitiveType === undefined) {
        return undefined;
    }
    return {
        isNullable: primitiveType.isNullable,
        isOptional: false,
        label: parameter.name,
        documentation: (_a = parameter.description) !== null && _a !== void 0 ? _a : undefined,
        typeKind: primitiveType.kind,
    };
}
function numberToCompletionItemKind(variant) {
    switch (variant) {
        case commonTypes_1.CompletionItemKind.Constant:
        case commonTypes_1.CompletionItemKind.Constructor:
        case commonTypes_1.CompletionItemKind.Enum:
        case commonTypes_1.CompletionItemKind.EnumMember:
        case commonTypes_1.CompletionItemKind.Event:
        case commonTypes_1.CompletionItemKind.Field:
        case commonTypes_1.CompletionItemKind.File:
        case commonTypes_1.CompletionItemKind.Folder:
        case commonTypes_1.CompletionItemKind.Function:
        case commonTypes_1.CompletionItemKind.Interface:
        case commonTypes_1.CompletionItemKind.Keyword:
        case commonTypes_1.CompletionItemKind.Method:
        case commonTypes_1.CompletionItemKind.Module:
        case commonTypes_1.CompletionItemKind.Operator:
        case commonTypes_1.CompletionItemKind.Property:
        case commonTypes_1.CompletionItemKind.Reference:
        case commonTypes_1.CompletionItemKind.Snippet:
        case commonTypes_1.CompletionItemKind.Struct:
        case commonTypes_1.CompletionItemKind.Text:
        case commonTypes_1.CompletionItemKind.TypeParameter:
        case commonTypes_1.CompletionItemKind.Unit:
        case commonTypes_1.CompletionItemKind.Value:
        case commonTypes_1.CompletionItemKind.Variable:
            return variant;
        default:
            return undefined;
    }
}
// Depending on which path we took to generate the symbols parameters that are lists will have one of two forms:
// 1. `list`
// 2. `list {$typeKind}`
// To make my life easier I'm going to allow both forms at this layer,
// and then normalize both forms to `list`.
function stringToPrimitiveType(text) {
    const split = text.split(" ");
    let isNullable;
    let typeKind;
    switch (split.length) {
        case 0:
            return undefined;
        case 1: {
            isNullable = false;
            typeKind = stringToTypeKind(text);
            break;
        }
        // Either `nullable $typeKind` or `list {$typeKind}`
        case 2: {
            const first = split[0];
            if (first === language_1.Constant.LanguageConstant.Nullable) {
                isNullable = true;
                typeKind = stringToTypeKind(split[1]);
                break;
            }
            else if (first === language_1.Constant.PrimitiveTypeConstant.List) {
                isNullable = false;
                typeKind = language_1.Type.TypeKind.List;
                break;
            }
            else {
                return undefined;
            }
        }
        // `nullable list {$typeKind}`
        case 3: {
            if (split[0] !== language_1.Constant.LanguageConstant.Nullable || split[1] !== language_1.Constant.PrimitiveTypeConstant.List) {
                return undefined;
            }
            isNullable = true;
            typeKind = language_1.Type.TypeKind.List;
            break;
        }
        default:
            return undefined;
    }
    return typeKind ? language_1.TypeUtils.primitiveType(isNullable, typeKind) : undefined;
}
function stringToTypeKind(text) {
    return language_1.ConstantUtils.isPrimitiveTypeConstant(text)
        ? language_1.TypeUtils.typeKindFromPrimitiveTypeConstantKind(text)
        : undefined;
}
//# sourceMappingURL=librarySymbolUtils.js.map