"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.LibraryProvider = void 0;
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const library_1 = require("../library");
const inspection_1 = require("../inspection");
const trace_1 = require("../trace");
class LibraryProvider {
    constructor(library, locale) {
        this.locale = locale;
        this.library = library;
        this.signatureInformationByLabel = new Map();
    }
    getAutocompleteItems(context) {
        return __awaiter(this, void 0, void 0, function* () {
            // eslint-disable-next-line require-await
            return yield powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
                var _a;
                const trace = context.traceManager.entry(trace_1.ProviderTraceConstant.LibrarySymbolProvider, this.getAutocompleteItems.name, context.initialCorrelationId);
                (_a = context.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
                if (!context.text || !context.range) {
                    trace.exit({ invalidContext: true });
                    return undefined;
                }
                const autocompleteItems = [];
                const contextText = context.text;
                for (const [label, definition] of this.library.libraryDefinitions.staticLibraryDefinitions.entries()) {
                    autocompleteItems.push(inspection_1.AutocompleteItemUtils.fromLibraryDefinition(label, definition, contextText));
                }
                for (const [label, definition] of this.library.libraryDefinitions.dynamicLibraryDefinitions().entries()) {
                    autocompleteItems.push(inspection_1.AutocompleteItemUtils.fromLibraryDefinition(label, definition, contextText));
                }
                trace.exit();
                return autocompleteItems;
            }), this.locale);
        });
    }
    getHover(context) {
        return __awaiter(this, void 0, void 0, function* () {
            // eslint-disable-next-line require-await
            return yield powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
                var _a, _b;
                const trace = context.traceManager.entry(trace_1.ProviderTraceConstant.LibrarySymbolProvider, this.getHover.name, context.initialCorrelationId);
                (_a = context.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
                if (!context.identifier) {
                    trace.exit({ invalidContext: true });
                    return undefined;
                }
                const identifierLiteral = context.identifier.literal;
                const definition = library_1.LibraryUtils.getDefinition(this.library, identifierLiteral);
                if (definition === undefined) {
                    trace.exit({ invalidContext: true });
                    return undefined;
                }
                const definitionText = LibraryProvider.getDefinitionKindText(definition.kind);
                const definitionTypeText = language_1.TypeUtils.nameOf(definition.asPowerQueryType, context.traceManager, trace.id);
                (_b = context.cancellationToken) === null || _b === void 0 ? void 0 : _b.throwIfCancelled();
                const hover = {
                    contents: {
                        kind: vscode_languageserver_types_1.MarkupKind.PlainText,
                        language: "powerquery",
                        value: `[${definitionText}] ${identifierLiteral}: ${definitionTypeText}`,
                    },
                    range: undefined,
                };
                trace.exit();
                return hover;
            }), this.locale);
        });
    }
    getSignatureHelp(context) {
        return __awaiter(this, void 0, void 0, function* () {
            // eslint-disable-next-line require-await
            return yield powerquery_parser_1.ResultUtils.ensureResultAsync(() => __awaiter(this, void 0, void 0, function* () {
                var _a, _b;
                const trace = context.traceManager.entry(trace_1.ProviderTraceConstant.LibrarySymbolProvider, this.getSignatureHelp.name, context.initialCorrelationId);
                (_a = context.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
                if (!context.functionName) {
                    trace.exit({ invalidContext: true });
                    return undefined;
                }
                const identifierLiteral = context.functionName;
                const definition = library_1.LibraryUtils.getDefinition(this.library, identifierLiteral);
                if (!library_1.LibraryDefinitionUtils.isFunction(definition)) {
                    trace.exit({ invalidContext: true });
                    return undefined;
                }
                const result = {
                    activeParameter: (_b = context.argumentOrdinal) !== null && _b !== void 0 ? _b : 0,
                    activeSignature: 0,
                    signatures: [this.getOrCreateSignatureInformation(identifierLiteral)],
                };
                trace.exit();
                return result;
            }), this.locale);
        });
    }
    static getDefinitionKindText(kind) {
        switch (kind) {
            case library_1.Library.LibraryDefinitionKind.Function:
                return "library function";
            case library_1.Library.LibraryDefinitionKind.Constant:
                return "library constant";
            case library_1.Library.LibraryDefinitionKind.Type:
                return "library type";
            default:
                throw powerquery_parser_1.Assert.isNever(kind);
        }
    }
    getOrCreateSignatureInformation(key) {
        if (!this.signatureInformationByLabel.has(key)) {
            const definition = library_1.LibraryDefinitionUtils.assertAsFunction(library_1.LibraryDefinitionUtils.getDefinition(this.library.libraryDefinitions, key));
            this.signatureInformationByLabel.set(key, library_1.LibraryUtils.signatureInformation(definition));
        }
        return powerquery_parser_1.Assert.asDefined(this.signatureInformationByLabel.get(key));
    }
}
exports.LibraryProvider = LibraryProvider;
//# sourceMappingURL=libraryProvider.js.map