"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.validateDuplicateIdentifiers = void 0;
const PQP = require("@microsoft/powerquery-parser");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const localization_1 = require("../localization");
const diagnosticErrorCode_1 = require("../diagnosticErrorCode");
const __1 = require("..");
const trace_1 = require("../trace");
function validateDuplicateIdentifiers(textDocument, nodeIdMapCollection, validationSettings, cancellationToken) {
    const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, validateDuplicateIdentifiers.name, validationSettings.initialCorrelationId);
    const updatedSettings = Object.assign(Object.assign({}, validationSettings), { initialCorrelationId: trace.id });
    if (!updatedSettings.checkForDuplicateIdentifiers) {
        trace.exit();
        return [];
    }
    const documentUri = textDocument.uri;
    const result = [
        ...validateDuplicateIdentifiersForLetExpresion(documentUri, nodeIdMapCollection, updatedSettings, trace.id, cancellationToken),
        ...validateDuplicateIdentifiersForRecord(documentUri, nodeIdMapCollection, updatedSettings, trace.id, cancellationToken),
        ...validateDuplicateIdentifiersForRecordType(documentUri, nodeIdMapCollection, updatedSettings, trace.id, cancellationToken),
        ...validateDuplicateIdentifiersForSection(documentUri, nodeIdMapCollection, updatedSettings, trace.id, cancellationToken),
    ];
    trace.exit();
    return result;
}
exports.validateDuplicateIdentifiers = validateDuplicateIdentifiers;
function validateDuplicateIdentifiersForLetExpresion(documentUri, nodeIdMapCollection, validationSettings, correlationId, cancellationToken) {
    var _a;
    const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, validateDuplicateIdentifiersForLetExpresion.name, correlationId);
    const letIds = (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.LetExpression)) !== null && _a !== void 0 ? _a : new Set();
    const result = validateDuplicateIdentifiersForKeyValuePair(documentUri, nodeIdMapCollection, [letIds], parser_1.NodeIdMapIterator.iterLetExpression, validationSettings, trace.id, cancellationToken);
    trace.exit();
    return result;
}
function validateDuplicateIdentifiersForRecord(documentUri, nodeIdMapCollection, validationSettings, correlationId, cancellationToken) {
    var _a, _b;
    const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, validateDuplicateIdentifiersForRecord.name, correlationId);
    const recordIds = [
        (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.RecordExpression)) !== null && _a !== void 0 ? _a : new Set(),
        (_b = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.RecordLiteral)) !== null && _b !== void 0 ? _b : new Set(),
    ];
    const result = validateDuplicateIdentifiersForKeyValuePair(documentUri, nodeIdMapCollection, recordIds, parser_1.NodeIdMapIterator.iterRecord, validationSettings, trace.id, cancellationToken);
    trace.exit();
    return result;
}
function validateDuplicateIdentifiersForRecordType(documentUri, nodeIdMapCollection, validationSettings, correlationId, cancellationToken) {
    var _a;
    const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, validateDuplicateIdentifiersForRecordType.name, correlationId);
    const recordTypeIds = [
        (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.RecordType)) !== null && _a !== void 0 ? _a : new Set(),
    ];
    const result = validateDuplicateIdentifiersForKeyValuePair(documentUri, nodeIdMapCollection, recordTypeIds, parser_1.NodeIdMapIterator.iterRecordType, validationSettings, trace.id, cancellationToken);
    trace.exit();
    return result;
}
function validateDuplicateIdentifiersForSection(documentUri, nodeIdMapCollection, validationSettings, correlationId, cancellationToken) {
    var _a;
    const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, validateDuplicateIdentifiersForSection.name, correlationId);
    const sectionIds = (_a = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.Section)) !== null && _a !== void 0 ? _a : new Set();
    const result = validateDuplicateIdentifiersForKeyValuePair(documentUri, nodeIdMapCollection, [sectionIds], parser_1.NodeIdMapIterator.iterSection, validationSettings, trace.id, cancellationToken);
    trace.exit();
    return result;
}
// Generalized logic for iterating over some collection and their children, essentially by doing:
//  for node in nodeIds:
//      for childOfNode in iterNodeFactory(node):
//          ...
function validateDuplicateIdentifiersForKeyValuePair(documentUri, nodeIdMapCollection, nodeIdCollections, iterNodeFactory, validationSettings, correlationId, cancellationToken) {
    const numIds = nodeIdCollections.reduce((partial, set) => partial + set.size, 0);
    const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, validateDuplicateIdentifiersForKeyValuePair.name, correlationId, { numIds });
    if (!numIds) {
        return [];
    }
    const result = [];
    for (const collection of nodeIdCollections) {
        for (const nodeId of collection) {
            cancellationToken === null || cancellationToken === void 0 ? void 0 : cancellationToken.throwIfCancelled();
            const node = parser_1.NodeIdMapUtils.assertXor(nodeIdMapCollection, nodeId);
            const duplicateFieldsByKey = new Map();
            const knownFieldByKey = new Map();
            for (const field of iterNodeFactory(nodeIdMapCollection, node)) {
                const keyLiteral = field.normalizedKeyLiteral;
                const duplicateFields = duplicateFieldsByKey.get(keyLiteral);
                const knownFeld = knownFieldByKey.get(keyLiteral);
                if (duplicateFields) {
                    duplicateFields.push(field);
                }
                else if (knownFeld) {
                    duplicateFieldsByKey.set(keyLiteral, [field, knownFeld]);
                }
                else {
                    knownFieldByKey.set(keyLiteral, field);
                }
            }
            for (const duplicates of duplicateFieldsByKey.values()) {
                const numFields = duplicates.length;
                const asRelatedInformation = duplicates.map((keyValuePair) => ({
                    location: {
                        uri: documentUri,
                        range: __1.PositionUtils.rangeFromTokenRange(keyValuePair.key.tokenRange),
                    },
                    message: createDuplicateIdentifierDiagnosticMessage(keyValuePair, validationSettings),
                }));
                for (let index = 0; index < numFields; index += 1) {
                    const duplicate = duplicates[index];
                    // Grab all DiagnosticRelatedInformation for a given key besides the one we're iterating over.
                    const relatedInformation = asRelatedInformation.filter((_, relatedIndex) => index !== relatedIndex);
                    result.push(createDuplicateIdentifierDiagnostic(duplicate, relatedInformation, validationSettings));
                }
            }
        }
    }
    trace.exit();
    return result;
}
function createDuplicateIdentifierDiagnostic(keyValuePair, relatedInformation, validationSettings) {
    return {
        code: diagnosticErrorCode_1.DiagnosticErrorCode.DuplicateIdentifier,
        message: createDuplicateIdentifierDiagnosticMessage(keyValuePair, validationSettings),
        range: __1.PositionUtils.rangeFromTokenRange(keyValuePair.key.tokenRange),
        relatedInformation,
        severity: vscode_languageserver_types_1.DiagnosticSeverity.Error,
        source: validationSettings.source,
    };
}
function createDuplicateIdentifierDiagnosticMessage(keyValuePair, validationSettings) {
    var _a;
    return localization_1.Localization.error_validation_duplicate_identifier(localization_1.LocalizationUtils.getLocalizationTemplates((_a = validationSettings === null || validationSettings === void 0 ? void 0 : validationSettings.locale) !== null && _a !== void 0 ? _a : PQP.DefaultLocale), keyValuePair.keyLiteral);
}
//# sourceMappingURL=validateDuplicateIdentifiers.js.map