"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.validateFunctionExpression = void 0;
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const localization_1 = require("../localization");
const diagnosticErrorCode_1 = require("../diagnosticErrorCode");
const __1 = require("..");
const trace_1 = require("../trace");
// Check for repeat parameter names for FunctionExpressions.
function validateFunctionExpression(validationSettings, nodeIdMapCollection) {
    var _a, _b;
    const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, validateFunctionExpression.name, validationSettings.initialCorrelationId);
    const updatedSettings = Object.assign(Object.assign({}, validationSettings), { initialCorrelationId: trace.id });
    const fnExpressionIds = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.FunctionExpression);
    if (fnExpressionIds === undefined) {
        trace.exit();
        return [];
    }
    const diagnostics = [];
    for (const nodeId of fnExpressionIds) {
        (_a = validationSettings.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
        diagnostics.push(validateNoDuplicateParameter(updatedSettings, nodeIdMapCollection, nodeId));
        (_b = updatedSettings.cancellationToken) === null || _b === void 0 ? void 0 : _b.throwIfCancelled();
    }
    trace.exit();
    return diagnostics.flat();
}
exports.validateFunctionExpression = validateFunctionExpression;
function validateNoDuplicateParameter(validationSettings, nodeIdMapCollection, fnExpressionId) {
    var _a, _b;
    const fnExpression = parser_1.NodeIdMapUtils.assertXorChecked(nodeIdMapCollection, fnExpressionId, language_1.Ast.NodeKind.FunctionExpression);
    const parameterNames = new Map();
    for (const parameter of parser_1.NodeIdMapIterator.iterFunctionExpressionParameterNames(nodeIdMapCollection, fnExpression)) {
        (_a = validationSettings.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
        const existingNames = (_b = parameterNames.get(parameter.literal)) !== null && _b !== void 0 ? _b : [];
        existingNames.push(parameter);
        parameterNames.set(parameter.literal, existingNames);
    }
    const diagnostics = [];
    for (const [identifierLiteral, nodes] of parameterNames) {
        if (nodes.length > 1) {
            diagnostics.push(...nodes.map((identifier) => {
                const parameterRange = __1.PositionUtils.rangeFromTokenRange(identifier.tokenRange);
                return createDiagnosticForDuplicateParameterName(validationSettings, parameterRange, identifierLiteral);
            }));
        }
    }
    return diagnostics;
}
function createDiagnosticForDuplicateParameterName(validationSettings, invokeExpressionRange, parameterName) {
    const templates = localization_1.LocalizationUtils.getLocalizationTemplates(validationSettings.locale);
    return {
        code: diagnosticErrorCode_1.DiagnosticErrorCode.DuplicateIdentifier,
        message: localization_1.Localization.error_validation_duplicate_identifier(templates, parameterName),
        range: invokeExpressionRange,
        severity: vscode_languageserver_types_1.DiagnosticSeverity.Error,
        source: validationSettings.source,
    };
}
//# sourceMappingURL=validateFunctionExpression.js.map