"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.validateInvokeExpression = void 0;
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const trace_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/common/trace");
const __1 = require("..");
const localization_1 = require("../localization");
const diagnosticErrorCode_1 = require("../diagnosticErrorCode");
const trace_2 = require("../trace");
function validateInvokeExpression(validationSettings, nodeIdMapCollection, typeCache) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const trace = validationSettings.traceManager.entry(trace_2.ValidationTraceConstant.Validation, validateInvokeExpression.name, validationSettings.initialCorrelationId);
        const updatedSettings = Object.assign(Object.assign({}, validationSettings), { initialCorrelationId: trace.id });
        const invokeExpressionIds = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.InvokeExpression);
        if (invokeExpressionIds === undefined) {
            trace.exit();
            return [];
        }
        const inspectionTasks = [];
        for (const nodeId of invokeExpressionIds) {
            (_a = validationSettings.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
            inspectionTasks.push(__1.Inspection.tryInvokeExpression(updatedSettings, nodeIdMapCollection, nodeId, typeCache));
        }
        const inspections = yield Promise.all(inspectionTasks);
        const diagnosticTasks = [];
        for (const triedInvokeExpression of inspections) {
            (_b = validationSettings.cancellationToken) === null || _b === void 0 ? void 0 : _b.throwIfCancelled();
            if (powerquery_parser_1.ResultUtils.isOk(triedInvokeExpression)) {
                diagnosticTasks.push(invokeExpressionToDiagnostics(updatedSettings, nodeIdMapCollection, triedInvokeExpression.value));
            }
            else {
                trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                throw triedInvokeExpression;
            }
        }
        const diagnostics = yield Promise.all(diagnosticTasks);
        trace.exit();
        return diagnostics.flat();
    });
}
exports.validateInvokeExpression = validateInvokeExpression;
function invokeExpressionToDiagnostics(validationSettings, nodeIdMapCollection, inspected) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const result = [];
        if (inspected.arguments !== undefined) {
            const invokeExpressionArguments = inspected.arguments;
            const givenArguments = inspected.arguments.givenArguments;
            const invokeExpressionRange = powerquery_parser_1.Assert.asDefined(yield __1.PositionUtils.rangeFromXorNode(nodeIdMapCollection, inspected.invokeExpressionXorNode), "expected at least one leaf node under InvokeExpression");
            const functionName = parser_1.NodeIdMapUtils.invokeExpressionIdentifierLiteral(nodeIdMapCollection, inspected.invokeExpressionXorNode.node.id);
            for (const [argIndex, mismatch] of invokeExpressionArguments.typeChecked.invalid.entries()) {
                (_a = validationSettings.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
                // eslint-disable-next-line no-await-in-loop
                const givenArgRange = yield __1.PositionUtils.rangeFromXorNode(nodeIdMapCollection, givenArguments[argIndex]);
                result.push(createDiagnosticForArgumentMismatch(validationSettings, mismatch, functionName, invokeExpressionRange, givenArgRange));
            }
            const numGivenArguments = invokeExpressionArguments.givenArguments.length;
            if (numGivenArguments < invokeExpressionArguments.numMinExpectedArguments ||
                numGivenArguments > invokeExpressionArguments.numMaxExpectedArguments) {
                result.push(createDiagnosticForArgumentNumberMismatch(validationSettings, invokeExpressionRange, invokeExpressionArguments.numMinExpectedArguments, invokeExpressionArguments.numMaxExpectedArguments, numGivenArguments));
            }
        }
        return result;
    });
}
function createDiagnosticForArgumentNumberMismatch(validationSettings, invokeExpressionRange, numMin, numMax, numGiven) {
    const templates = localization_1.LocalizationUtils.getLocalizationTemplates(validationSettings.locale);
    return {
        code: diagnosticErrorCode_1.DiagnosticErrorCode.InvokeArgumentCountMismatch,
        message: localization_1.Localization.error_validation_invokeExpression_numArgs(templates, numMin, numMax, numGiven),
        range: invokeExpressionRange,
        severity: vscode_languageserver_types_1.DiagnosticSeverity.Error,
        source: validationSettings.source,
    };
}
function createDiagnosticForArgumentMismatch(validationSettings, mismatch, functionName, invokeExpressionRange, givenArgRange) {
    var _a;
    let range;
    let message;
    const parameter = mismatch.expected;
    const argName = parameter.nameLiteral;
    const expected = language_1.TypeUtils.nameOfTypeKind((_a = parameter.type) !== null && _a !== void 0 ? _a : language_1.Type.AnyInstance.kind);
    // An argument containing at least one leaf node was given.
    if (givenArgRange) {
        const actual = language_1.TypeUtils.nameOfTypeKind(powerquery_parser_1.Assert.asDefined(mismatch.actual).kind);
        range = givenArgRange;
        message = createTypeMismatchMessage(validationSettings.locale, functionName, argName, expected, actual);
    }
    else {
        range = invokeExpressionRange;
        message = createMissingMandatoryMessage(validationSettings.locale, functionName, argName);
    }
    return {
        code: diagnosticErrorCode_1.DiagnosticErrorCode.InvokeArgumentTypeMismatch,
        message,
        range,
        severity: vscode_languageserver_types_1.DiagnosticSeverity.Error,
        source: validationSettings.source,
    };
}
function createMissingMandatoryMessage(locale, functionName, argName) {
    const templates = localization_1.LocalizationUtils.getLocalizationTemplates(locale);
    return localization_1.Localization.error_validation_invokeExpression_missingMandatory(templates, functionName, argName);
}
function createTypeMismatchMessage(locale, functionName, argName, expected, actual) {
    const templates = localization_1.LocalizationUtils.getLocalizationTemplates(locale);
    return localization_1.Localization.error_validation_invokeExpression_typeMismatch(templates, functionName, argName, expected, actual);
}
//# sourceMappingURL=validateInvokeExpression.js.map