"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.validateUnknownIdentifiers = void 0;
const powerquery_parser_1 = require("@microsoft/powerquery-parser");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const parser_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/parser");
const language_1 = require("@microsoft/powerquery-parser/lib/powerquery-parser/language");
const __1 = require("..");
const localization_1 = require("../localization");
const jaroWinkler_1 = require("../jaroWinkler");
const diagnosticErrorCode_1 = require("../diagnosticErrorCode");
const externalType_1 = require("../externalType/externalType");
const trace_1 = require("../trace");
function validateUnknownIdentifiers(validationSettings, nodeIdMapCollection, typeCache) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, validateUnknownIdentifiers.name, validationSettings.initialCorrelationId);
        const updatedSettings = Object.assign(Object.assign({}, validationSettings), { initialCorrelationId: trace.id });
        // Grab all identifiers in the given context.
        const identifierValues = findIdentifierAsts(nodeIdMapCollection);
        // For each identifier create a Promise<TriedScope>, then await on all.
        const identifiersWithNodeScope = yield powerquery_parser_1.ArrayUtils.mapAsync(identifierValues, (identifier) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            (_a = validationSettings.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
            const identifierExpression = parser_1.NodeIdMapUtils.parentAstChecked(nodeIdMapCollection, identifier.id, language_1.Ast.NodeKind.IdentifierExpression);
            const literal = identifierExpression && identifierExpression.inclusiveConstant !== undefined
                ? `@${identifier.literal}`
                : identifier.literal;
            return {
                identifier: identifierExpression !== null && identifierExpression !== void 0 ? identifierExpression : identifier,
                literal,
                triedNodeScope: yield __1.Inspection.tryNodeScope(updatedSettings, nodeIdMapCollection, updatedSettings.eachScopeById, identifier.id, typeCache.scopeById),
            };
        }));
        const unknownIdentifiers = findUnknownIdentifiers(validationSettings, identifiersWithNodeScope, trace.id);
        const result = unknownIdentifiersToDiagnostics(updatedSettings, unknownIdentifiers);
        trace.exit();
        return result;
    });
}
exports.validateUnknownIdentifiers = validateUnknownIdentifiers;
// For context, "Tbl.AsdC" has a score of ~0.59 against "Table.AddColumn"
const JaroWinklerSuggestionThreshold = 0.9;
function findIdentifierAsts(nodeIdMapCollection) {
    const identifiers = nodeIdMapCollection.idsByNodeKind.get(language_1.Ast.NodeKind.Identifier);
    if (identifiers === undefined) {
        return [];
    }
    const astNodeById = nodeIdMapCollection.astNodeById;
    const result = [];
    for (const identifierId of identifiers.values()) {
        const identifier = astNodeById.get(identifierId);
        if (identifier !== undefined && identifier.identifierContextKind === language_1.Ast.IdentifierContextKind.Value) {
            result.push(identifier);
        }
    }
    return result;
}
function findUnknownIdentifiers(validationSettings, identifiersWithNodeScope, correlationId) {
    const trace = validationSettings.traceManager.entry(trace_1.ValidationTraceConstant.Validation, findUnknownIdentifiers.name, correlationId);
    const unknownIdentifiers = [];
    for (const identifierWithNodeScope of identifiersWithNodeScope) {
        if (powerquery_parser_1.ResultUtils.isError(identifierWithNodeScope.triedNodeScope)) {
            continue;
        }
        const identifier = identifierWithNodeScope.identifier;
        const nodeScope = identifierWithNodeScope.triedNodeScope.value;
        const literal = identifierWithNodeScope.literal;
        if (!nodeScope.has(literal) &&
            !(literal[0] === "@" && nodeScope.has(literal.slice(1))) &&
            !__1.LibraryUtils.hasDefinition(validationSettings.library, literal) &&
            // even no external type found
            !validationSettings.library.externalTypeResolver({
                kind: externalType_1.ExternalTypeRequestKind.Value,
                identifierLiteral: literal,
            })) {
            const knownIdentifiers = [
                ...nodeScope.keys(),
                ...__1.LibraryUtils.getDefinitionKeys(validationSettings.library),
            ];
            const [jaroWinklerScore, suggestion] = (0, jaroWinkler_1.calculateJaroWinklers)(literal, knownIdentifiers);
            unknownIdentifiers.push({
                identifier,
                literal,
                suggestion: jaroWinklerScore > JaroWinklerSuggestionThreshold ? suggestion : undefined,
            });
        }
    }
    trace.exit();
    return unknownIdentifiers;
}
function unknownIdentifiersToDiagnostics(validationSettings, unknownIdentifiers) {
    const templates = localization_1.LocalizationUtils.getLocalizationTemplates(validationSettings.locale);
    return unknownIdentifiers.map((unknownIdentifier) => ({
        code: diagnosticErrorCode_1.DiagnosticErrorCode.UnknownIdentifier,
        message: localization_1.Localization.error_validation_unknownIdentifier(templates, unknownIdentifier.literal, unknownIdentifier.suggestion),
        range: __1.PositionUtils.rangeFromTokenRange(unknownIdentifier.identifier.tokenRange),
        severity: vscode_languageserver_types_1.DiagnosticSeverity.Error,
        source: validationSettings.source,
    }));
}
//# sourceMappingURL=validateUnknownIdentifiers.js.map